#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""API to use to submit jobs in the ILC VO.

:since:  Apr 13, 2010
:author: Stephane Poss
"""

from __future__ import absolute_import
import os

from DIRAC.Interfaces.API.Dirac import Dirac
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC import gConfig, S_ERROR, S_OK, gLogger

from ILCDIRAC.Core.Utilities.ProcessList import ProcessList

__RCSID__ = "$Id$"

# pylint: disable=protected-access

COMPONENT_NAME = 'DiracILC'
LOG = gLogger.getSubLogger(__name__)


class DiracILC(Dirac):
  """DiracILC is VO specific API Dirac.

  Adding specific ILC functionalities to the Dirac class, and implement the :func:`preSubmissionChecks` method
  """

  def __init__(self, withRepo=False, repoLocation=''):
    """Internal initialization of the ILCDIRAC API."""
    super(DiracILC, self).__init__(withRepo, repoLocation)
    self.software_versions = {}
    self.checked = False
    self.processList = None
    self.ops = Operations()

  def getProcessList(self):
    """Get the :mod:`~ILCDIRAC.Core.Utilities.ProcessList.ProcessList` needed by :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.Whizard`.

    :return: process list object
    """
    processlistpath = gConfig.getValue("/LocalSite/ProcessListPath", "")
    if not processlistpath:
      LOG.info('Will download the process list locally. To gain time, please put it somewhere and add to \
      your dirac.cfg the entry /LocalSite/ProcessListPath pointing to the file')
      pathtofile = self.ops.getValue("/ProcessList/Location", "")
      if not pathtofile:
        LOG.error("Could not get path to process list")
        processlist = ""
      else:
        datMan = DataManager()
        datMan.getFile(pathtofile)
        processlist = os.path.basename(pathtofile)
    else:
      processlist = processlistpath
    self.processList = ProcessList(processlist)
    return self.processList

  def preSubmissionChecks(self, job, mode=None):
    """Overridden method from :mod:`DIRAC.Interfaces.API.Dirac`

    Checks from CS that required software packages are available.

    :param job: job instance
    :type job: ~ILCDIRAC.Interfaces.API.NewInterface.Job.Job
    :param mode: submission mode, not used here.

    :return: :func:`~DIRAC.Core.Utilities.ReturnValues.S_OK` , :func:`~DIRAC.Core.Utilities.ReturnValues.S_ERROR`
    """

    if not job.oktosubmit:
      LOG.error('You should use job.submit(dirac)')
      return S_ERROR("You should use job.submit(dirac)")
    res = self._do_check(job)
    if not res['OK']:
      return res
    if not self.checked:
      res = job._askUser()
      if not res['OK']:
        return res
      self.checked = True
    return S_OK()

  def checkparams(self, job):
    """Helper method.

    Method used for stand alone checks of job integrity. Calls the formulation error checking of the job

    Actually checks that all input are available and checks that the required software packages are available in the CS

    :param job: Job Instance
    :type job: ~ILCDIRAC.Interfaces.API.NewInterface.Job.Job
    :return: :func:`~DIRAC.Core.Utilities.ReturnValues.S_OK` , :func:`~DIRAC.Core.Utilities.ReturnValues.S_ERROR`
    """
    formulationErrors = job.errorDict

    if formulationErrors:
      for method, errorList in formulationErrors.items():
        LOG.error('>>>> Error in %s() <<<<\n%s' % (method, '\n'.join(errorList)))
      return S_ERROR(formulationErrors)
    return self.preSubmissionChecks(job, mode='')

  def giveProcessList(self):
    """Returns the list of Processes."""
    return self.processList

  def retrieveRepositoryOutputDataLFNs(self, requestedStates=None):
    """Helper function.

    Get the list of uploaded output data for a set of jobs in a repository

    :param requestedStates: List of states requested for filtering the list
    :type requestedStates: list of strings
    :return: list
    """
    if requestedStates is None:
      requestedStates = ['Done']
    llist = []
    if not self.jobRepo:
      LOG.warn("No repository is initialised")
      return S_OK()
    jobs = self.jobRepo.readRepository()['Value']
    for jobID in sorted(jobs.keys()):
      jobDict = jobs[jobID]
      if 'State' in jobDict and (jobDict['State'] in requestedStates):
        if ('UserOutputData' in jobDict and (not int(jobDict['UserOutputData']))) or \
           ('UserOutputData' not in jobDict):
          params = self.getJobParameters(int(jobID))
          if params['OK']:
            if 'UploadedOutputData' in params['Value']:
              lfn = params['Value']['UploadedOutputData']
              llist.append(lfn)
    return llist

  def _do_check(self, job):
    """Main method for checks.

    :param job: :mod:`job object <ILCDIRAC.Interfaces.API.NewInterface.Job.Job>`
    :return: :func:`~DIRAC.Core.Utilities.ReturnValues.S_OK` , :func:`~DIRAC.Core.Utilities.ReturnValues.S_ERROR`
    """
    # Start by taking care of sandbox
    if hasattr(job, "inputsandbox") and isinstance(job.inputsandbox, list) and len(job.inputsandbox):
      found_list = False
      for items in job.inputsandbox:
        if isinstance(items, list):  # We fix the SB in the case is contains a list of lists
          found_list = True
          for inBoxFile in items:
            if isinstance(inBoxFile, list):
              return S_ERROR("Too many lists of lists in the input sandbox, please fix!")
            job.inputsandbox.append(inBoxFile)
          job.inputsandbox.remove(items)
      resolvedFiles = job._resolveInputSandbox(job.inputsandbox)
      if found_list:
        LOG.warn("Input Sandbox contains list of lists. Please avoid that.")
      fileList = ';'.join(resolvedFiles)
      description = 'Input sandbox file list'
      job._addParameter(job.workflow, 'InputSandbox', 'JDL', fileList, description)

    res = self.checkInputSandboxLFNs(job)
    if not res['OK']:
      return res

    platform = job.workflow.findParameter("Platform").getValue()
    apps = job.workflow.findParameter("SoftwarePackages")
    if apps:
      apps = apps.getValue()
      for appver in apps.split(";"):
        app = appver.split(".")[0].lower()  # first element
        vers = appver.split(".")[1:]  # all the others
        vers = '.'.join(vers)
        res = self._checkapp(platform, app, vers)
        if not res['OK']:
          return res
    outputpathparam = job.workflow.findParameter("UserOutputPath")
    if outputpathparam:
      outputpath = outputpathparam.getValue()
      res = self._checkoutputpath(outputpath)
      if not res['OK']:
        return res
    useroutputdata = job.workflow.findParameter("UserOutputData")
    useroutputsandbox = job.addToOutputSandbox
    if useroutputdata:
      res = self._checkdataconsistency(useroutputdata.getValue(), useroutputsandbox)
      if not res['OK']:
        return res

    return S_OK()

  def _checkapp(self, platform, appName, appVersion):
    """Check availability of application in CS.

    :param str platform: System platform
    :param str appName: Application name
    :param str appVersion: Application version
    :return: :func:`~DIRAC.Core.Utilities.ReturnValues.S_OK` or :func:`~DIRAC.Core.Utilities.ReturnValues.S_ERROR`
    """
    csPathTarBall = "/AvailableTarBalls/%s/%s/%s/TarBall" % (platform, appName, appVersion)
    csPathCVMFS = "/AvailableTarBalls/%s/%s/%s/CVMFSPath" % (platform, appName, appVersion)

    LOG.debug("Checking for software version in " + csPathTarBall)
    app_version = self.ops.getValue(csPathTarBall, '')

    LOG.debug("Checking for software version in " + csPathCVMFS)
    app_version_cvmfs = self.ops.getValue(csPathCVMFS, '')

    if not app_version and not app_version_cvmfs:
      LOG.error("Could not find the specified software %s_%s for %s, check in CS" % (appName, appVersion, platform))
      return S_ERROR(
          "Could not find the specified software %s_%s for %s, check in CS" %
          (appName, appVersion, platform))
    return S_OK()

  def _checkoutputpath(self, path):
    """Validate the outputpath specified for the application.

    :param str path: Path of output data
    :return: :func:`~DIRAC.Core.Utilities.ReturnValues.S_OK` , :func:`~DIRAC.Core.Utilities.ReturnValues.S_ERROR`
    """
    if path.find("//") > -1 or path.find("/./") > -1 or path.find("/../") > -1:
      LOG.error("OutputPath of setOutputData() contains invalid characters, please remove any //, /./, or /../")
      return S_ERROR("Invalid path")
    path = path.rstrip()
    if path[-1] == "/":
      LOG.error("Please strip trailing / from outputPath in setOutputData()")
      return S_ERROR("Invalid path")
    return S_OK()

  def _checkdataconsistency(self, useroutputdata, useroutputsandbox):
    """Make sure the files are either in OutpuSandbox or OutputData but not both.

    :param list useroutputdata: List of files set in the outputdata
    :param list useroutputsandbox: List of files set in the output sandbox
    :return: :func:`~DIRAC.Core.Utilities.ReturnValues.S_OK` , :func:`~DIRAC.Core.Utilities.ReturnValues.S_ERROR`
    """
    useroutputdata = useroutputdata.split(";")
    for data in useroutputdata:
      for item in useroutputsandbox:
        if data == item:
          LOG.error("Output data and sandbox should not contain the same things.")
          return S_ERROR("Output data and sandbox should not contain the same things.")
      if data.find("*") > -1:
        LOG.error("Remove wildcard characters from output data definition: must be exact files")
        return S_ERROR("Wildcard character in OutputData definition")
    return S_OK()

  def checkInputSandboxLFNs(self, job):
    """Check that LFNs in the InputSandbox exist in the FileCatalog.

    :param job: :mod:`job object <ILCDIRAC.Interfaces.API.NewInterface.Job.Job>`
    :return: :func:`~DIRAC.Core.Utilities.ReturnValues.S_OK` , :func:`~DIRAC.Core.Utilities.ReturnValues.S_ERROR`
    """
    lfns = []
    inputsb = job.workflow.findParameter("InputSandbox")
    if inputsb:
      isblist = inputsb.getValue()
      if isblist:
        isblist = isblist.split(';')
        for inBoxFile in isblist:
          if inBoxFile.lower().count('lfn:'):
            lfns.append(inBoxFile.replace('LFN:', '').replace('lfn:', ''))

    if not lfns:
      return S_OK()

    res = self.getReplicas(lfns)
    if not res["OK"]:
      return S_ERROR('Could not get replicas')
    failed = res['Value']['Failed']
    if failed:
      LOG.error('Failed to find replicas for the following files %s' % ', '.join(failed))
      return S_ERROR('Failed to find replicas')

    LOG.info('All LFN files have replicas available')
    singleReplicaSEs = set(self.ops.getValue("/UserJobs/InputSandbox/SingleReplicaSEs", []))
    preferredSEs = set(self.ops.getValue("/UserJobs/InputSandbox/PreferredSEs", []))
    minimumNumberOfReplicas = self.ops.getValue("/UserJobs/InputSandbox/MinimumNumberOfReplicas", 2)
    failSubmission = []
    for lfn, replicas in res['Value']['Successful'].items():
      sites = set(replicas.keys())
      if singleReplicaSEs.intersection(sites):
        continue
      if len(replicas) < minimumNumberOfReplicas:
        LOG.error("ERROR: File %r has less than %s replicas,"
                       "please use dirac-dms-replicate-lfn to replicate to e.g.,:%s"
                       % (lfn, minimumNumberOfReplicas, ", ".join(preferredSEs - set(replicas.keys()))))
        failSubmission.append(lfn)
        for site in set((preferredSEs - set(replicas.keys()))):
          LOG.error("  dirac-dms-replicate-lfn %s %s" % (lfn, site))
        LOG.error("Or use job.setInputData for data files")
    if failSubmission:
      return S_ERROR("Not enough replicas for %s" % ",".join(failSubmission))

    return S_OK()

  def submitJob(self, job, mode='wms'):
    """Submit jobs to DIRAC WMS. These can be either:

        - Instances of the Job Class
           - VO Application Jobs
           - Inline scripts
           - Scripts as executables
           - Scripts inside an application environment

        - JDL File
        - JDL String

       Example usage:

       >>> print dirac.submitJob(job)
       {'OK': True, 'Value': '12345'}

       :param job: Instance of Job class or JDL string
       :type job: `ILCDIRAC.Interfaces.API.NewInterface.Job.Job` or str
       :param mode: Submit job locally with mode = ``'wms'`` (default), ``'local'`` to run workflow or ``'agent'`` to
           run full Job Wrapper locally
       :type mode: str
       :returns: S_OK,S_ERROR
    """
    LOG.debug("Submitting job")
    return super(DiracILC, self).submitJob(job, mode)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Base application class. All applications inherit from this class.

The interfaces for the currently available applications can be found here:

.. hlist::
   :columns: 3

   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.CheckCollections`,
   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.CheckWNs`,
   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.GenericApplication`,
   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.GetSRMFile`,
   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.LCSIM`,
   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.Marlin`,
   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.Mokka`,
   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.OverlayInput`,
   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.PostGenSelection`,
   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.Pythia`,
   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.RootMacro`,
   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.RootScript`,
   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.SLCIOConcatenate`,
   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.SLCIOSplit`,
   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.SLICPandora`,
   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.SLIC`,
   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.StdhepCutJava`,
   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.StdhepCut`,
   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.StdHepSplit`,
   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.Tomato`,
   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.Whizard`
   * :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.DDSim`

:author: Stephane Poss
:author: Remi Ete
:author: Ching Bon Lam
"""

from __future__ import absolute_import
import inspect
import os
import sys
import types
import six.moves.urllib.request
import six.moves.urllib.parse
import six.moves.urllib.error

from DIRAC.Core.Workflow.Module import ModuleDefinition
from DIRAC.Core.Workflow.Parameter import Parameter

from DIRAC import S_OK, S_ERROR, gLogger
import six

LOG = gLogger.getSubLogger(__name__)


__RCSID__ = "$Id$"

# pylint: disable=no-self-use, unused-argument


class Application(object):  # pylint: disable=too-many-instance-attributes
  """General application definition.

  Any new application should inherit from this class.
  """
  # need to define slots
  ## __slots__ = []

  def __init__(self, paramdict=None):
    """Can define the full application by passing a dictionary in the constructor.

    >>> app = Application({"appname":"marlin","version":"v0111Prod",
    ...                    "steeringFile":"My_file.xml","numberOfEvents":1000})

    :param dict paramdict: Dictionary of parameters that can be set. Will throw an exception if one of them does not exist.
    """
    super(Application, self).__init__()
    # application nane (executable)
    self.appname = ""
    # application version
    self.version = ""
    # Steering file (duh!)
    self.steeringFile = ""
    # Input sandbox: steering file automatically added to SB
    self.inputSB = []
    # Input file
    self.inputFile = ""
    # Output file
    self.outputFile = ""
    self.outputPath = ""
    self.outputDstFile = ''
    self.outputDstPath = ''
    self.outputRecFile = ''
    self.outputRecPath = ''
    self.outputSE = ''
    self._listofoutput = []
    # Log file
    self.logFile = ""

    # Detector type (ILD or SID)
    self.detectortype = ""
    # Data type : gen, SIM, REC, DST
    self.datatype = ""

    # Extra command line arguments
    self.extraCLIArguments = ""

    # Needed for Generation+Stdhepcut
    self.willBeCut = False

    # Debug mode
    self.debug = False

    # Prod Parameters: things that appear on the prod details
    self.prodparameters = {}
    self.accountInProduction = True
    # Module name and description: Not to be set by the users, internal call only, used to get the Module objects
    self._modulename = ''
    self._moduledescription = ''
    self._importLocation = "ILCDIRAC.Workflow.Modules"

    # System Configuration: comes from Job definition
    self._platform = ''

    # Internal member: hold the list of the job's application set before self: used when using getInputFromApp
    self._job = None
    self._jobapps = []
    self._jobsteps = []
    self._jobtype = ''
    # input application: will link the OutputFile of the guys in there with the InputFile of the self
    self._inputapp = []
    self._linkedidx = None
    # Needed to link the parameters.
    self._inputappstep = None

    # flag set to true in Job.append
    self.addedtojob = False
    # Following are needed for error report
    self._errorDict = {}

    # This is used to filter out the members that should not be set when using a dict as input
    self._paramsToExclude = ['_paramsToExclude', "_log", "_errorDict", "addedtojob", '_ops',
                             "_inputappstep", "_linkedidx", "_inputapp", "_jobtype",
                             "_jobsteps", "_jobapps", "_job", "_platform", "_importLocation",
                             "_moduledescription", "_modulename", "prodparameters",
                             "datatype", "detectortype", "_listofoutput", "inputSB",
                             "appname", 'accountInProduction', 'outputPath']

    # Next is to use the setattr method.
    self._setparams(paramdict)

  def __repr__(self):
    classstr = "%s" % self.appname
    if self.version:
      classstr += " %s" % self.version
    return classstr

  def _setparams(self, params):
    """Call the setter that was passed in the input dictionary."""
    if not params:
      return S_OK()
    for param, value in params.items():
      try:
        getattr(self, "set%s" % param)(value)
      except AttributeError:
        LOG.error("The %s class does not have a set%s method." % (self.__class__.__name__, param))
    return S_OK()

  def _getParamsDict(self):
    """Return dictionary that can be used to build a new application based on the current."""
    curdict = self.__dict__
    pdict = {}
    for key, val in curdict.items():
      if key not in self._paramsToExclude:
        if val:
          pdict[key] = val
    return S_OK(pdict)

  def setName(self, name):
    """Define name of application.

    :param str name: Name of the application. Normally, every application defines its own, so no need to call that one
    """
    self._checkArgs({'name': (str,)})
    self.appname = name
    return S_OK()

  def setVersion(self, version):
    """Define version to use.

    :param str version: Version of the application to use
    """
    self._checkArgs({'version': (str,)})
    self.version = version
    return S_OK()

  def setSteeringFile(self, steeringfile):
    """Set the steering file, and add it to the sandbox.

    :param str steeringfile: Steering file to use. Can be any type: whizard.in, mokka.steer, slic.mac, marlin.xml, lcsim.lcsim, etc.
    """
    self._checkArgs({'steeringfile': (str,)})
    self.steeringFile = steeringfile
    if os.path.exists(steeringfile) or steeringfile.lower().count("lfn:"):
      self.inputSB.append(steeringfile)
    return S_OK()

  def setLogFile(self, logfile):
    """Define application log file.

    :param str logfile: Log file to use. Set by default if not set.
    """
    self._checkArgs({'logfile': (str,)})
    self.logFile = logfile
    return S_OK()

  def setOutputFile(self, ofile, path=None):
    """Set the output file.

    :param str ofile: Output file name. Will overwrite the default. This is
       necessary when linking applications (when using :any:`getInputFromApp`)

    :param str path: Set the output path for the output file to go. Will not
       do anything in a :mod:`~ILCDIRAC.Interfaces.API.NewInterface.UserJob`. Use
       :func:`~ILCDIRAC.Interfaces.API.NewInterface.UserJob.UserJob.setOutputData`
       of the job for that functionality.
    """
    self._checkArgs({'ofile': (str,)})

    self.outputFile = ofile
    self.prodparameters[ofile] = {}
    if self.detectortype:
      self.prodparameters[ofile]['detectortype'] = self.detectortype
    if self.datatype:
      self.prodparameters[ofile]['datatype'] = self.datatype

    if path:
      self._checkArgs({'path': (str,)})
      self.outputPath = path

    return S_OK()

  def setOutputSE(self, se):
    """Set the output storage element for all files produced by this application.

    :param str se: Storage element name. Example CERN-SRM, IN2P3-SRM, RAL-SRM, IMPERIAL-SRM
    """
    self._checkArgs({'se': (str,)})
    self.outputSE = se
    return S_OK()

  def setInputFile(self, inputfile):
    """Set the input file to use: stdhep, slcio, root, whatever.

    :param inputfile: Input file (data, not steering) to pass to the application. Can be local file or LFN:
    :type inputfile: string or list
    """
    kwargs = {"inputfile": inputfile}
    if not isinstance(inputfile, (six.string_types, list)):
      return self._reportError("InputFile must be string or list of strings", __name__, **kwargs)
    if not isinstance(inputfile, list):
      inputfile = [inputfile]
    for inf in inputfile:
      if os.path.exists(inf) or inf.lower().count("lfn:"):
        self.inputSB.append(inf)

    self.inputFile = ";".join(inputfile)

    return S_OK()

  def getInputFromApp(self, application):
    """Called to link applications.

    >>> mokka = Mokka()
    >>> marlin = Marlin()
    >>> marlin.getInputFromApp(mokka)

    :param application: Application to link against.
    :type application: :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Application`
    """
    self._inputapp.append(application)
    return S_OK()

  def setDebug(self, debug=True):
    """Set the application to debug mode.

    >>> app = Application()
    >>> app.setDebug()

    :param bool debug: Set the application to debug mode. Default is True when called. If not, then it's false.
    """
    self._checkArgs({"debug": bool})
    self.debug = debug
    return S_OK()

  def setExtraCLIArguments(self, arguments):
    """Pass any command line arguments as a string to the application.

    :param str arguments:
    """
    self._checkArgs({"arguments": (str,)})
    self.extraCLIArguments = arguments
    return S_OK()

  def listAttributes(self):
    """Method to list attributes for users.

    Doesn't list any private or semi-private attributes
    """
    LOG.notice('Attribute list :')
    for key, val in self.__dict__.items():
      if key not in self._paramsToExclude:
        if not val:
          val = "Not defined"
        LOG.notice("  %s: %s" % (key, val))

  def checkProductionMetaData(self, metaDict):
    """check that the production metadata coming out of this application is correct.

    Implemented in applications where this is necessary. Called from `Job.append`.

    :param dict metaDict: production job metadata dictionary, will be updated
    :returns: S_OK, S_ERROR
    """
    LOG.debug("Checking production metadata for application %s " % self.__class__.__name__)
    for key, val in metaDict.items():
      LOG.debug("%s: %s" % (key, val))

    return S_OK()


########################################################################################
#    More private methods: called by the applications of the jobs, but not by the users
#    Please, do not touch when you don't know what you are doing.
#                           / \  //\
#            |\___/|      /   \//  \\
#            /0  0  \__  /    //  | \ \
#           /     /  \/_/    //   |  \  \
#           @_^_@'/   \/_   //    |   \   \
#           //_^_/     \/_ //     |    \    \
#        ( //) |        \///      |     \     \
#      ( / /) _|_ /   )  //       |      \     _\
#    ( // /) '/,_ _ _/  ( ; -.    |    _ _\.-~        .-~~~^-.
#  (( / / )) ,-{        _      `-.|.-~-.           .~         `.
# (( // / ))  '/\      /                 ~-. _ .-~      .-~^-.  \
# (( /// ))      `.   {            }                   /      \  \
#  (( / ))     .----~-.\        \-'                 .~         \  `. \^-.
#             ///.----..>        \             _ -~             `.  ^-`  ^-_
#               ///-._ _ _ _ _ _ _}^ - - - - ~                     ~-- ,.-~
#                                                                  /.-~
########################################################################################

  def _setApplicationModuleAndParameters(self, stepdefinition):
    """Create Application Module, add it to a Step and set values to Module.

    Called in every applications
    """
    m1 = self._applicationModule()
    stepdefinition.addModule(m1)
    m1i = stepdefinition.createModuleInstance(m1.getType(), stepdefinition.getType())
    self._applicationModuleValues(m1i)
    return S_OK()

  def _setUserJobFinalization(self, stepdefinition):
    """Create UserOutputDataModule and add it to Step.

    Called after the private method setApplicationModuleAndParameters in some user job applications
    """
    m2 = self._getUserOutputDataModule()
    stepdefinition.addModule(m2)
    stepdefinition.createModuleInstance(m2.getType(), stepdefinition.getType())
    return S_OK()

  def _setOutputComputeDataList(self, stepdefinition):
    """Create ComputeOutputDataListModule and add it to Step.

    Called after the private method setApplicationModuleAndParameters in some production job applications
    """
    m2 = self._getComputeOutputDataListModule()
    stepdefinition.addModule(m2)
    stepdefinition.createModuleInstance(m2.getType(), stepdefinition.getType())
    return S_OK()

  def _createModuleDefinition(self):
    """Create Module definition.

    As it's generic code, all apps will use this.
    """
    moduledefinition = ModuleDefinition(self._modulename)
    moduledefinition.setDescription(self._moduledescription)
    body = 'from %s.%s import %s\n' % (self._importLocation, self._modulename, self._modulename)
    moduledefinition.setBody(body)
    return moduledefinition

  def _getUserOutputDataModule(self):
    """This is separated as not all applications require user specific output data (i.e. GetSRMFile and Overlay).

    Only used in UserJobs.

    The UserJobFinalization only runs last. It's called every step, but is running only if last.
    """
    moduledefinition = ModuleDefinition('UserJobFinalization')
    moduledefinition.setDescription('Uploads user output data files with specific policies.')
    body = 'from %s.%s import %s\n' % (self._importLocation, 'UserJobFinalization', 'UserJobFinalization')
    moduledefinition.setBody(body)
    return moduledefinition

  def _getComputeOutputDataListModule(self):
    """This is separated from the applications as this is used in production jobs only."""
    moduledefinition = ModuleDefinition("ComputeOutputDataList")
    moduledefinition.setDescription("Compute the output data list to be treated by the last finalization")
    body = 'from %s.%s import %s\n' % (self._importLocation, "ComputeOutputDataList", "ComputeOutputDataList")
    moduledefinition.setBody(body)
    return moduledefinition

  def _applicationModule(self):
    """Create the module for the application, and add the parameters to it.

    Overloaded by every application class.
    """
    raise NotImplementedError('Must be implemented in Application')

  def _applicationModuleValues(self, moduleinstance):
    """Set the values for the modules parameters.

    """
    pass

  def _userjobmodules(self, stepdefinition):
    """Method used to return the needed module for UserJobs.

    It's different from the ProductionJobs
    (userJobFinalization for instance)
    """
    LOG.error("This application does not implement the modules, you get an empty list")
    return S_ERROR('Not implemented')

  def _prodjobmodules(self, stepdefinition):
    """Same as above, but the other way around."""
    LOG.error("This application does not implement the modules, you get an empty list")
    return S_ERROR('Not implemented')

  def _checkConsistency(self, job=None):
    """Called from Job Class, overloaded by every class. Used to check that everything is fine, in particular that all required parameters are defined.

    :param job: The instance of the job
    :type job: :func:`~ILCDIRAC.Interfaces.API.NewInterface.Job.Job`
    :returns: S_OK/S_ERROR
    """
    return S_OK()

  def _checkFinalConsistency(self):
    """Called from Job Class, overloaded by every class.

    Used to check that everything is fine, in particular that
    all required parameters are defined.
    Some info are passed from the job to the applications: this is then used to check that it makes the app valid
    """
    return S_OK()

  def _checkWorkflowConsistency(self):
    """Called from Job Class, overloaded by every class.

    Used to check the workflow consistency: linking between
    applications
    Should also call :any:`_checkRequiredApp` when needed.
    """
    return S_OK()

  def _checkRequiredApp(self):
    """Called by :any:`_checkWorkflowConsistency` when relevant."""
    if self._inputapp:
      for app in self._inputapp:
        if app not in self._jobapps:
          return S_ERROR("job order not correct: If this app uses some input coming from an other app, the app in \
          question must be passed to job.append() before.")
        else:
          self._linkedidx = self._jobapps.index(app)

    return S_OK()

  def _addBaseParameters(self, stepdefinition):
    """Add to step the default parameters.

    appname, version, steeringfile, (nbevts, Energy), LogFile, InputFile, OutputFile, OutputPath.
    """
    stepdefinition.addParameter(Parameter("applicationName", "", "string", "", "", False, False,
                                          "Application Name"))
    stepdefinition.addParameter(Parameter("applicationVersion", "", "string", "", "", False, False,
                                          "Application Version"))
    stepdefinition.addParameter(Parameter("SteeringFile", "", "string", "", "", False, False, "Steering File"))
    stepdefinition.addParameter(Parameter("applicationLog", "", "string", "", "", False, False, "Log File"))
    stepdefinition.addParameter(
        Parameter(
            "ExtraCLIArguments",
            "",
            "string",
            "",
            "",
            False,
            False,
            "Extra CLI arguments"))
    stepdefinition.addParameter(Parameter("InputFile", "", "string", "", "", True, False, "Input File"))

    if len(self.outputFile):
      stepdefinition.addParameter(Parameter("OutputFile", "", "string", "", "", False, False, "Output File"))
    if len(self.outputDstFile):
      stepdefinition.addParameter(Parameter("outputDST", "", "string", "", "", False, False,
                                            "Output DST File"))
    if len(self.outputRecFile):
      stepdefinition.addParameter(Parameter("outputREC", "", "string", "", "", False, False,
                                            "Output REC File"))

    stepdefinition.addParameter(Parameter("OutputPath", "", "string", "", "", True, False,
                                          "Output File path on the grid"))
    stepdefinition.addParameter(Parameter("outputPathREC", "", "string", "", "", True, False,
                                          "Output REC File path on the grid"))
    stepdefinition.addParameter(Parameter("outputPathDST", "", "string", "", "", True, False,
                                          "Output DST File path on the grid"))
    stepdefinition.addParameter(Parameter("OutputSE", "", "string", "", "", True, False,
                                          "Output File storage element"))
    stepdefinition.addParameter(Parameter('listoutput', [], "list", "", "", False, False,
                                          "list of output file name"))
    # Following should be workflow parameters
    # stepdefinition.addParameter(Parameter("NbOfEvents",         0,    "int", "", "", False, False,
    #                                      "Number of events to process"))
    #stepdefinition.addParameter(Parameter("Energy",             0,    "int", "", "", False, False, "Energy"))

    return self._getSpecificAppParameters(stepdefinition)

  def _getSpecificAppParameters(self, stepdef):
    """Add specific parameters, should be overloaded."""
    return S_OK()

  def _setBaseStepParametersValues(self, stepinstance):
    """Set the values for the basic step parameters."""

    stepinstance.setValue("applicationName", self.appname)
    stepinstance.setValue("applicationVersion", self.version)
    stepinstance.setValue("applicationLog", self.logFile)
    stepinstance.setValue("SteeringFile", self.steeringFile)
    if not self._inputapp:
      stepinstance.setValue("InputFile", self.inputFile)

    if len(self.outputFile):
      stepinstance.setValue("OutputFile", self.outputFile)
    if len(self.outputRecFile):
      stepinstance.setValue("outputREC", self.outputRecFile)
    if len(self.outputDstFile):
      stepinstance.setValue("outputDST", self.outputDstFile)
    stepinstance.setValue("OutputPath", self.outputPath)
    stepinstance.setValue("outputPathREC", self.outputRecPath)
    stepinstance.setValue("outputPathDST", self.outputDstPath)
    stepinstance.setValue("OutputSE", self.outputSE)
    stepinstance.setValue('listoutput', self._listofoutput)
    stepinstance.setValue('ExtraCLIArguments', six.moves.urllib.parse.quote(self.extraCLIArguments))
    return self._setSpecificAppParameters(stepinstance)

  def _setSpecificAppParameters(self, stepinstance):
    """Set the value of the parameters.

    Should be overloaded
    """
    return S_OK()

  def _addParametersToStep(self, stepdefinition):
    """Method to be overloaded by every application.

    Add the parameters to the given step.
    Should call any:`_addBaseParameters`.
    Called from Job
    """
    return self._addBaseParameters(stepdefinition)

  def _setStepParametersValues(self, stepinstance):
    """Method to be overloaded by every application.

    For all parameters that are not to be linked,
    set the values in the step instance
    Called from Job
    """
    return self._setBaseStepParametersValues(stepinstance)

  def _resolveLinkedStepParameters(self, stepinstance):
    """Method to be overloaded by every application that resolve what are the linked parameters (e.g. OuputFile and InputFile).

    See :class:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.StdhepCut.StdhepCut` for example.
    Called from Job.
    """
    return S_OK()

  def _analyseJob(self, job):
    """Called from Job, only gives the application the knowledge of the Job (application, step, system config)"""
    self._job = job

    self._platform = job.workflow.findParameter("Platform").getValue()

    self._jobapps = job.applicationlist

    self._jobsteps = job.steps

    self._jobtype = job.type

    return self._doSomethingWithJob()

  def _doSomethingWithJob(self):
    """ As name suggest, if there is something to do with the job, it should be done now
    Example: software to install
    """
    return S_OK()

  def _addedtojob(self):
    """Called from Job to tell the application that it is added to job and the parameter values are set.

    Prevents the user from thinking he can change the values after the app was added to job.
    """
    self.addedtojob = True

  def _checkArgs(self, argNamesAndTypes):
    """Private method to check the validity of the parameters."""

    # inspect.stack()[1][0] returns the frame object ([0]) of the caller
    # function (stack()[1]).
    # The frame object is required for getargvalues. Getargvalues returns
    # a tuple with four items. The fourth item ([3]) contains the local
    # variables in a dict.

    args = inspect.getargvalues(inspect.stack()[1][0])[3]

    #

    for argName, argType in argNamesAndTypes.items():

      if argName not in args:
        self._reportError('Method does not contain argument \'%s\'' % argName,
                           __name__,
                           **self._getArgsDict(1)
                         )

      if not isinstance(args[argName], argType):
        self._reportError('Argument \'%s\' is not of type %s' % (argName, argType),
                           __name__,
                           **self._getArgsDict(1)
                         )

  def _getArgsDict(self, level=0):
    """Private method."""

    # Add one to stack level such that we take the caller function as the
    # reference point for 'level'

    level += 1

    #

    args = inspect.getargvalues(inspect.stack()[level][0])
    adict = {}

    for arg in args[0]:

      if arg == "self":
        continue

      # args[3] contains the 'local' variables

      adict[arg] = args[3][arg]

    return adict

  #############################################################################
  def _reportError(self, message, short_name='', **kwargs):
    """Internal Function.

    Gets caller method name and arguments, formats the
    information and adds an error to the global error dictionary to be
    returned to the user.
    Stolen from DIRAC Job Class
    """
    className = short_name
    if not short_name:
      className = __name__
    methodName = sys._getframe(1).f_code.co_name
    arguments = []
    for key in kwargs:
      if kwargs[key]:
        arguments.append('%s = %s ( %s )' % (key, kwargs[key], type(kwargs[key])))
    finalReport = 'Problem with %s.%s() call:\nArguments: %s\nMessage: %s\n' % (className, methodName,
                                                                                 ', '.join(arguments),
                                                                                 message)
    if methodName in self._errorDict:
      tmp = self._errorDict[methodName]
      tmp.append(finalReport)
      self._errorDict[methodName] = tmp
    else:
      self._errorDict[methodName] = [finalReport]
    LOG.error(finalReport)
    return S_ERROR(finalReport)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test the Core Confogiration."""

from __future__ import absolute_import
import unittest
from mock import MagicMock as Mock
from ILCDIRAC.Core.Utilities import Configuration

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Core.Utilities.Configuration'


class TestCheckConf(unittest.TestCase):
  """Test the Configuration Utilities Module."""

  def setUp(self):
    pass

  def tearDown(self):
    pass

  def test_getOptionValue(self):
    ops = Mock()

    def getValueMock(path, defVal):
      return {"/base/myOp": "baseVal",
              "/base/foo/myOp": "fooVal",
              "/base/foo/bar/myOp": "barVal",
              "/base/foo/bar/baz/myOp": "bazVal",
             }.get(path, defVal)

    ops.getValue.side_effect = getValueMock

    value = Configuration.getOptionValue(ops, "/base", "myOp", "defVal", ['foo', 'bar', 'baz'])
    self.assertEqual(value, 'bazVal')

    value = Configuration.getOptionValue(ops, "/base", "myOp", "defVal", ['foo', 'bar', 'baz2'])
    self.assertEqual(value, 'barVal')

    value = Configuration.getOptionValue(ops, "/base", "myOp", "defVal", ['foo', 'bar2', 'baz'])
    self.assertEqual(value, 'fooVal')

    value = Configuration.getOptionValue(ops, "/base", "myOp", "defVal", ['args', '', 'kwargs'])
    self.assertEqual(value, 'baseVal')

    value = Configuration.getOptionValue(ops, "/base2", "myOp", "defVal", ['args', '', 'kwargs'])
    self.assertEqual(value, 'defVal')



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Create a list of LFNs form a repository file created during job submission.

The repository file is defined when creating the :class:`~ILCDIRAC.Interfaces.API.DiracILC.DiracILC` instance.

Options:

  -r repoLocation       Path to repository file


:since: Apr 22, 2010
:author: Stephane Poss
"""
from __future__ import absolute_import
__RCSID__ = "$Id$"

from DIRAC.Core.Base import Script
from DIRAC.Core.Utilities.DIRACScript import DIRACScript
from DIRAC import exit as dexit
from DIRAC import S_OK, gLogger
LOG = gLogger.getSubLogger(__name__)


class _Params(object):
  """dummy."""

  def __init__(self):
    self.repo = ''

  def setRepo(self, optionVal):
    self.repo = optionVal
    return S_OK()

  def registerSwitches(self):
    Script.registerSwitch('r:', 'repository=', 'Path to repository file', self.setRepo)
    Script.setUsageMessage('\n'.join([__doc__.split('\n')[1],
                                        '\nUsage:',
                                        '  %s [option|cfgfile] ...\n' % Script.scriptName]))


@DIRACScript()
def main():
  """create the LFnList."""
  cliparams = _Params()
  cliparams.registerSwitches()
  Script.parseCommandLine(ignoreErrors=False)

  repoLocation = cliparams.repo
  if not repoLocation:
    Script.showHelp()
    dexit(2)
  from ILCDIRAC.Interfaces.API.DiracILC import DiracILC
  dirac = DiracILC(True, repoLocation)

  dirac.monitorRepository(False)
  lfns = []
  lfns = dirac.retrieveRepositoryOutputDataLFNs()
  LOG.notice("lfnlist=[")
  for lfn in lfns:
    LOG.notice('"LFN:%s",' % lfn)
  LOG.notice("]")
  dexit(0)


if __name__ == "__main__":
  main()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Unit tests for the FccAnalysis.py file."""

from __future__ import absolute_import
import unittest
import os
from mock import patch, MagicMock as Mock
from ILCDIRAC.Workflow.Modules.FccAnalysis import FccAnalysis
from Tests.Utilities.GeneralUtils import assertEqualsImproved, \
    assertDiracFailsWith, assertDiracSucceedsWith

from DIRAC import S_OK, S_ERROR

import six


__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Workflow.Modules.FccAnalysis'
BUILTIN_NAME = 'builtins' if six.PY3 else '__builtin__'


class TestFccAnalysis(unittest.TestCase):
  """Test the FccAnalysis module."""

  def setUp(self):
    """set up the objects."""

    def replace_realpath(path):
      return os.path.join("/test/realpath", path)

    self.log_mock = Mock()

    self.patches = [patch("%s.os.path.realpath" % MODULE_NAME, new=Mock(side_effect=replace_realpath)),
                    patch("%s.os.path.dirname" % MODULE_NAME, new=Mock(return_value="/test/dirname")),
                    patch("%s.FccAnalysis.redirectLogOutput" % MODULE_NAME, new=Mock()),
                    patch('%s.LOG' % MODULE_NAME, new=self.log_mock),
                    ]

    for patcher in self.patches:
      patcher.start()

    self.fccAna = FccAnalysis()

    self.fccAna.platform = "Testplatform123"
    self.fccAna.applicationLog = "testlog123"
    self.fccAna.SteeringFile = os.path.realpath("fccConfFile.cfg")
    self.fccAna.applicationName = "fccApp"
    self.fccAna.applicationVersion = "v1.0"
    self.fccAna.STEP_NUMBER = "1"
    self.fccAppIndex = "%s_%s_Step_%s" % (self.fccAna.applicationName, self.fccAna.applicationVersion,
                                          self.fccAna.STEP_NUMBER)
    self.applicationScript = os.path.realpath("%s.sh" % self.fccAppIndex)
    self.root_files = ["outputFile.root", "outputFile2.root"]
    self.exists_dict = {
        self.fccAna.SteeringFile: True,
        self.fccAna.applicationLog: True,
        "/test/realpath/outputFile.root": True,
        '/test/realpath/outputFile.txt': True
        }

  def replace_exists(self, path):
    return self.exists_dict[path]

  def tearDown(self):
    for patcher in self.patches:
      patcher.stop()
    del self.fccAna

  def test_runit_noplatform(self):
    self.fccAna.platform = None
    assertDiracFailsWith(self.fccAna.runIt(), 'No ILC platform selected', self)

  def test_runit_noapplog(self):
    self.fccAna.platform = "Testplatform123"
    self.fccAna.applicationLog = None
    assertDiracFailsWith(self.fccAna.runIt(), 'No Log file provided', self)

  def test_runit_workflowbad(self):
    self.fccAna.applicationLog = "testlog123"
    self.fccAna.workflowStatus = {'OK': False}
    assertDiracSucceedsWith(self.fccAna.runIt(), 'should not proceed', self)

  def test_runit_stepbad(self):
    self.fccAna.stepStatus = {'OK': False}
    assertDiracSucceedsWith(self.fccAna.runIt(), 'should not proceed', self)

  @patch("%s.getEnvironmentScript" % MODULE_NAME, new=Mock(return_value=S_OK()))
  @patch("%s.os.path.exists" % MODULE_NAME, new=Mock(return_value=True))
  def test_getenvironmentscript_cfg_lookup(self):
    self.assertTrue(self.fccAna.getEnvironmentScript())

  @patch("%s.getEnvironmentScript" % MODULE_NAME, new=Mock(return_value=S_ERROR()))
  def test_getenvironmentscript_cfg_lookup_failed(self):
    self.assertFalse(self.fccAna.getEnvironmentScript())
    self.log_mock.error.assert_called_once_with("Environment : 'dirac.cfg' file look up failed")

  @patch("%s.getEnvironmentScript" % MODULE_NAME, new=Mock(return_value=S_OK()))
  @patch("%s.os.path.exists" % MODULE_NAME, new=Mock(return_value=False))
  def test_getenvironmentscript_pathexists_failed(self):
    self.assertFalse(self.fccAna.getEnvironmentScript())

  def test_writetofile(self):
    data_to_write = 'LEAVE ME HERE'
    with patch('%s.open' % BUILTIN_NAME, create=True) as mock_write:
      self.assertTrue(self.fccAna.writeToFile("w", "/file/to/write", data_to_write))

      mock_write.assert_called_once_with("/file/to/write", 'w')
      manager = mock_write.return_value.__enter__.return_value
      manager.write.assert_called_with(data_to_write)

      debug_message = "Application : File write operation successfull"
      self.log_mock.debug.assert_called_once_with(debug_message)

  @patch('%s.open' % MODULE_NAME, new=Mock(side_effect=IOError("ioerror")), create=True)
  def test_writetofile_failed(self):
    self.assertFalse(self.fccAna.writeToFile("w", "", ""))

    error_message = "Application : File write operation failed\nioerror"
    self.log_mock.error.assert_called_once_with(error_message)

  def test_generatescriptonthefly(self):
    error_message = (
        "Environment : Environment script not found\n"
        "for this configuration : conf, name, version\n"
        "Can not generate one dynamically"
        )
    assertDiracFailsWith(self.fccAna.generateScriptOnTheFly("conf", "name", "version"), error_message, self)
    # This is the only one error message put in debug level, check FccAnalysis
    self.log_mock.debug.assert_called_once_with(error_message)

  def test_generatebashscript(self):
    with patch('%s.FccAnalysis.writeToFile' % MODULE_NAME, new=Mock(return_value=True)) as mock_write, \
        patch('%s.os.chmod' % MODULE_NAME) as mock_chmod:

      self.assertTrue(self.fccAna.generateBashScript("command1"))
      mock_write.assert_called_once_with('w', self.fccAna.applicationScript, '#!/bin/bash\nsource \ncommand1\n')
      mock_chmod.assert_called_once_with(self.fccAna.applicationScript, 0o755)
      self.log_mock.debug.assert_any_call("Application code : Bash script creation successfull")
      self.log_mock.debug.assert_any_call("Application file : Bash script rights setting successfull")

  def test_generatebashscript_write_failed(self):
    with patch('%s.FccAnalysis.writeToFile' % MODULE_NAME, new=Mock(return_value=False)) as mock_write:

      self.assertFalse(self.fccAna.generateBashScript("command1"))

      mock_write.assert_called_once_with('w', self.fccAna.applicationScript, '#!/bin/bash\nsource \ncommand1\n')
      error_message = "Application code : Bash script creation failed"
      self.log_mock.error.assert_called_once_with(error_message)

  @patch("%s.FccAnalysis.writeToFile" % MODULE_NAME, new=Mock(return_value=True))
  def test_generategaudiconffile(self):
    self.fccAna.logLevel = "DEBUG"
    self.assertTrue(self.fccAna.generateGaudiConfFile())

  def test_generategaudiconffile_gaudioptions(self):
    self.fccAna.logLevel = "DEBUG"
    self.fccAna.NumberOfEvents = 42
    self.fccAna.RandomSeed = 126
    self.fccAna.randomGenerator["Gaudi"] = True
    self.fccAna.read = True
    self.fccAna.InputData = ["/path/to/data"]
    self.fccAna.InputFile = []

    gaudiOptions = ["from Configurables import ApplicationMgr"]
    gaudiOptions += ["from Gaudi.Configuration import *"]

    eventSetting = ["# N) AUTOMATIC GENERATION OF CODE DONE BY FCC APPLICATION FOR EVENT NUMBER SETTING"]
    eventSetting += ["ApplicationMgr().EvtMax=%s" % self.fccAna.NumberOfEvents]
    gaudiOptions += eventSetting

    seedSetting = ["# N) AUTOMATIC GENERATION OF CODE DONE BY FCC APPLICATION FOR SEED NUMBER SETTING"]
    seedSetting += ["from Configurables import SimG4Svc, RndmGenSvc"]
    seedSetting += ['from GaudiSvc.GaudiSvcConf import HepRndm__Engine_CLHEP__RanluxEngine_']
    seedSetting += ["randomEngine = eval('HepRndm__Engine_CLHEP__RanluxEngine_')"]
    seedSetting += ["randomEngine = randomEngine('RndmGenSvc.Engine')"]
    seedSetting += ["randomEngine.Seeds = [%d]  " % self.fccAna.RandomSeed]

    gaudiOptions += seedSetting

    levelSetting = "ApplicationMgr().OutputLevel=%s" % self.fccAna.logLevel
    gaudiOptions += [levelSetting]

    fccswPodioOptions = ["# N) AUTOMATIC GENERATION OF CODE DONE BY FCC APPLICATION FOR INPUT FILE SETTING"]
    fccswPodioOptions += ["from Configurables import FCCDataSvc"]
    fccswPodioOptions += ["import os"]

    fccInputDataSubstitution = ['%s' for data in self.fccAna.InputData]
    fccInputData = ["os.path.realpath(os.path.basename('%s'))" % data
                    for data in self.fccAna.InputData]
    # We can provide many input files to FCCDataSvc() like this :
    inputSetting = [
        "podioevent = FCCDataSvc('EventDataSvc', input='%s' %% (%s))" %
        (" ".join(fccInputDataSubstitution), ", ".join(fccInputData))]
    inputSetting += ["ApplicationMgr().ExtSvc += [podioevent]"]
    fccswPodioOptions += inputSetting

    gaudiOptions += fccswPodioOptions

    with patch("%s.FccAnalysis.writeToFile" % MODULE_NAME) as mock_write:
      mock_write.return_value = True
      self.assertTrue(self.fccAna.generateGaudiConfFile())
      mock_write.assert_called_once_with('w', self.fccAna.gaudiOptionsFile, "\n".join(gaudiOptions) + '\n')

  def test_generategaudiconffile_levelfailed(self):
    self.fccAna.logLevel = "GUBED"
    self.assertFalse(self.fccAna.generateGaudiConfFile())
    message = (
        "FCCSW specific consistency : Invalid value for the log level\n"
        "Possible values for the log level are :\n%(log)s" % {'log': " ".join(self.fccAna.logLevels)}
        )
    self.log_mock.error.assert_called_with(message)

  @patch("%s.FccAnalysis.writeToFile" % MODULE_NAME, new=Mock(return_value=False))
  def test_generategaudiconffile_writefailed(self):
    self.fccAna.logLevel = "DEBUG"
    self.assertFalse(self.fccAna.generateGaudiConfFile())

  @patch('%s.FccAnalysis.getEnvironmentScript' % MODULE_NAME, new=Mock(return_value=False))
  def test_runit_getenvironmentscript_failed(self):
    error_message = (
        "Environment : Environment script look up failed\n"
        "Failed to get environment"
        )
    assertDiracFailsWith(self.fccAna.runIt(), error_message, self)
    self.log_mock.error.assert_called_once_with(error_message)

  @patch('%s.FccAnalysis.getEnvironmentScript' % MODULE_NAME, new=Mock(return_value=True))
  @patch("%s.os.path.exists" % MODULE_NAME, new=Mock(return_value=False))
  def test_runit_conffile_failed(self):
    error_message = (
        "Environment : FCC configuration file '%(sfile)s' does not exist,"
        " can not run FCC application" % {'sfile': self.fccAna.SteeringFile}
        )
    assertDiracFailsWith(self.fccAna.runIt(), error_message, self)
    self.log_mock.info.assert_any_call("Environment : Environment script look up successfull")
    self.log_mock.error.assert_called_once_with(error_message)

  @patch('%s.FccAnalysis.getEnvironmentScript' % MODULE_NAME, new=Mock(return_value=True))
  @patch('%s.FccAnalysis.generateGaudiConfFile' % MODULE_NAME, new=Mock(return_value=False))
  def test_runit_generategaudiconffile_failed(self):
    self.fccAna.isGaudiOptionsFileNeeded = True

    with patch('os.path.exists') as mock_exists:

      mock_exists.side_effect = self.replace_exists
      error_message = "Application code : generateGaudiConfFile() failed"
      assertDiracFailsWith(self.fccAna.runIt(), error_message, self)
      self.log_mock.error.assert_called_once_with(error_message)
      self.log_mock.info.assert_any_call("Environment : Environment script look up successfull")

  @patch('%s.FccAnalysis.getEnvironmentScript' % MODULE_NAME, new=Mock(return_value=True))
  @patch("%s.FccAnalysis.generateBashScript" % MODULE_NAME, new=Mock(return_value=False))
  def test_runit_generatebashscript_failed(self):
    with patch('os.path.exists') as mock_exists:

      mock_exists.side_effect = self.replace_exists
      error_message = "Application code : Creation of the bash script failed"
      assertDiracFailsWith(self.fccAna.runIt(), error_message, self)
      self.log_mock.info.assert_any_call("Environment : Environment script look up successfull")
      self.log_mock.error.assert_called_once_with(error_message)

  @patch('%s.FccAnalysis.getEnvironmentScript' % MODULE_NAME, new=Mock(return_value=True))
  @patch("%s.FccAnalysis.generateBashScript" % MODULE_NAME, new=Mock(return_value=True))
  def test_runit_shellcall_failed(self):
    with patch('os.path.exists') as mock_exists, \
        patch("%s.shellCall" % MODULE_NAME) as mock_shellcall:

      mock_exists.side_effect = self.replace_exists
      mock_shellcall.return_value = S_ERROR()

      error_message = "Application : Application execution failed"
      assertDiracFailsWith(self.fccAna.runIt(), error_message, self)
      self.log_mock.info.assert_any_call("Environment : Environment script look up successfull")
      self.log_mock.error.assert_called_once_with(error_message)
      mock_shellcall.assert_called_once_with(
          0,
          self.fccAna.applicationScript,
          callbackFunction=self.fccAna.redirectLogOutput,
          bufferLimit=20971520)
      self.log_mock.debug.assert_any_call("Application code : Creation of the bash script successfull")
      self.log_mock.debug.assert_any_call("Application : Application execution and log file creation...")

  @patch('%s.FccAnalysis.getEnvironmentScript' % MODULE_NAME, new=Mock(return_value=True))
  @patch("%s.FccAnalysis.generateBashScript" % MODULE_NAME, new=Mock(return_value=True))
  @patch('%s.glob.glob' % MODULE_NAME, new=Mock(return_value=[]))
  def test_runit_log_failed_withoutapp(self):
    self.fccAna.ignoreapperrors = True
    self.exists_dict[self.fccAna.applicationLog] = False

    with patch('os.makedirs'), \
         patch('os.path.exists') as mock_exists, \
         patch("%s.shellCall" % MODULE_NAME, new=Mock(return_value={'OK': True, 'Value': ["", "stdout", "stderr"]})) as mock_shellcall:

      mock_exists.side_effect = self.replace_exists
      assertDiracSucceedsWith(self.fccAna.runIt(), "Execution of the FCC application successfull", self)
      self.log_mock.info.assert_any_call("Environment : Environment script look up successfull")
      self.log_mock.debug.assert_any_call("Application code : Creation of the bash script successfull")
      self.log_mock.debug.assert_any_call("Application : Application execution and log file creation...")
      self.log_mock.debug.assert_any_call("Application : Application execution successfull")

      self.log_mock.error.assert_called_once_with("Application : Log file creation failed")
      mock_shellcall.assert_called_once_with(
          0,
          self.fccAna.applicationScript,
          callbackFunction=self.fccAna.redirectLogOutput,
          bufferLimit=20971520)

  @patch('%s.FccAnalysis.getEnvironmentScript' % MODULE_NAME, new=Mock(return_value=True))
  @patch("%s.FccAnalysis.generateBashScript" % MODULE_NAME, new=Mock(return_value=True))
  def test_runit_log_failed_withapp(self):
    self.exists_dict[self.fccAna.applicationLog] = False
    self.fccAna.ignoreapperrors = False

    with patch('os.makedirs'), \
         patch('os.path.exists') as mock_exists, \
         patch("%s.shellCall" % MODULE_NAME, new=Mock(return_value={'OK': True, 'Value': ["", "stdout", "stderr"]})) as mock_shellcall:

      mock_exists.side_effect = self.replace_exists
      error_message = '%s did not produce the expected log %s' % (
          self.fccAna.applicationName, self.fccAna.applicationLog)
      assertDiracFailsWith(self.fccAna.runIt(), error_message, self)
      self.log_mock.error.assert_any_call(error_message)
      self.log_mock.info.assert_any_call("Environment : Environment script look up successfull")
      self.log_mock.debug.assert_any_call("Application code : Creation of the bash script successfull")
      self.log_mock.debug.assert_any_call("Application : Application execution and log file creation...")
      self.log_mock.debug.assert_any_call("Application : Application execution successfull")
      mock_shellcall.assert_called_once_with(
          0,
          self.fccAna.applicationScript,
          callbackFunction=self.fccAna.redirectLogOutput,
          bufferLimit=20971520)

  @patch('%s.FccAnalysis.getEnvironmentScript' % MODULE_NAME, new=Mock(return_value=True))
  @patch("%s.FccAnalysis.generateBashScript" % MODULE_NAME, new=Mock(return_value=True))
  @patch('%s.FccAnalysis.writeToFile' % MODULE_NAME, new=Mock(return_value=True))
  @patch("%s.glob.glob" % MODULE_NAME, new=Mock(return_value=[]))
  def test_runit(self):
    with patch('os.makedirs'), \
        patch('os.path.exists') as mock_exists, \
        patch("%s.shellCall" % MODULE_NAME, new=Mock(return_value={'OK': True, 'Value': ["", "stdout", "stderr"]})) as mock_shellcall:

      mock_exists.side_effect = self.replace_exists

      assertDiracSucceedsWith(self.fccAna.runIt(), "Execution of the FCC application successfull", self)
      self.log_mock.info.assert_any_call("Environment : Environment script look up successfull")
      self.log_mock.debug.assert_any_call("Application code : Creation of the bash script successfull")
      self.log_mock.debug.assert_any_call("Application : Application execution and log file creation...")
      self.log_mock.debug.assert_any_call("Application : Application execution successfull")
      self.log_mock.debug.assert_any_call("Application : Log file creation successfull")
      self.log_mock.warn.assert_called_once_with(
          "Application : no root files have been generated, was that intentional ?")
      mock_shellcall.assert_called_once_with(
          0,
          self.fccAna.applicationScript,
          callbackFunction=self.fccAna.redirectLogOutput,
          bufferLimit=20971520)

  @patch('%s.FccAnalysis.getEnvironmentScript' % MODULE_NAME, new=Mock(return_value=True))
  @patch("%s.FccAnalysis.generateBashScript" % MODULE_NAME, new=Mock(return_value=True))
  @patch('%s.FccAnalysis.writeToFile' % MODULE_NAME, new=Mock(return_value=True))
  @patch('%s.glob.glob' % MODULE_NAME, new=Mock(return_value=[]))
  def test_runit_with_inputfile(self):
    get_input_from_app = "output_%s.root" % (self.fccAna.fccAppIndex)
    self.fccAna.step_commons['InputFile'] = get_input_from_app

    input_file = os.path.realpath(get_input_from_app)

    self.exists_dict[input_file] = True

    with patch('os.makedirs'), \
         patch('os.path.exists') as mock_exists:

      mock_exists.side_effect = self.replace_exists
      assertDiracSucceedsWith(self.fccAna.runIt(), "Execution of the FCC application successfull", self)

      assertEqualsImproved(self.fccAna.InputFile, input_file, self)
      assertEqualsImproved(self.fccAna.SteeringFile, self.fccAna.InputFile, self)
      self.log_mock.debug.assert_any_call(
          "Application : Configuration file taken from the input file(s) '%s'" %
          self.fccAna.InputFile)

  @patch('%s.FccAnalysis.getEnvironmentScript' % MODULE_NAME, new=Mock(return_value=True))
  @patch("%s.FccAnalysis.generateBashScript" % MODULE_NAME, new=Mock(return_value=True))
  @patch('%s.FccAnalysis.writeToFile' % MODULE_NAME, new=Mock(return_value=True))
  @patch('%s.glob.glob' % MODULE_NAME, new=Mock(return_value=[]))
  def test_runit_with_many_inputfiles(self):
    get_input_from_app = ["output1_%s.root" % self.fccAppIndex, "output2_%s.root" % self.fccAppIndex]
    self.fccAna.step_commons['InputFile'] = get_input_from_app

    input_file1 = os.path.realpath(get_input_from_app[0])
    input_file2 = os.path.realpath(get_input_from_app[1])

    self.exists_dict[input_file1] = True
    self.exists_dict[input_file2] = True

    with patch('os.makedirs'), \
         patch('os.path.exists') as mock_exists:

      mock_exists.side_effect = self.replace_exists
      assertDiracSucceedsWith(self.fccAna.runIt(), "Execution of the FCC application successfull", self)

      assertEqualsImproved(self.fccAna.InputFile, [input_file1, input_file2], self)
      assertEqualsImproved(self.fccAna.SteeringFile, " ".join(self.fccAna.InputFile), self)
      self.log_mock.debug.assert_any_call(
          "Application : Configuration file taken from the input file(s) '%s'" %
          self.fccAna.SteeringFile)

  @patch('%s.FccAnalysis.getEnvironmentScript' % MODULE_NAME, new=Mock(return_value=True))
  @patch("%s.FccAnalysis.generateBashScript" % MODULE_NAME, new=Mock(return_value=True))
  @patch('%s.FccAnalysis.writeToFile' % MODULE_NAME, new=Mock(return_value=True))
  @patch('%s.glob.glob' % MODULE_NAME, new=Mock(return_value=[]))
  def test_runit_without_inputfile(self):
    with patch('os.makedirs'), \
        patch('os.path.exists') as mock_exists:

      mock_exists.side_effect = self.replace_exists
      assertDiracSucceedsWith(self.fccAna.runIt(), "Execution of the FCC application successfull", self)
      assertEqualsImproved(self.fccAna.InputFile, [], self)

  @patch('%s.FccAnalysis.getEnvironmentScript' % MODULE_NAME, new=Mock(return_value=True))
  @patch("%s.FccAnalysis.generateBashScript" % MODULE_NAME, new=Mock(return_value=True))
  @patch('%s.FccAnalysis.writeToFile' % MODULE_NAME, new=Mock(return_value=True))
  def test_runit_without_current_rootfiles(self):
    getctime_dict = {
        os.path.realpath(self.root_files[1]): 1501667507.9749944,
        os.path.realpath(self.root_files[0]): 1501667510.7510207,
        os.path.realpath(self.applicationScript): 1501667512.7510207
        }

    def replace_getctime(path):
      return getctime_dict[os.path.realpath(path)]

    with patch('os.makedirs'), \
         patch("%s.glob.glob" % MODULE_NAME) as mock_glob, \
         patch("os.path.getctime") as mock_getctime, \
         patch('os.path.exists') as mock_exists:

      mock_exists.side_effect = self.replace_exists
      mock_getctime.side_effect = replace_getctime
      mock_glob.return_value = self.root_files
      assertDiracSucceedsWith(self.fccAna.runIt(), "Execution of the FCC application successfull", self)

      self.log_mock.warn.assert_called_once_with(
          "Application : This application did not generate any root files, was that intentional ?")

  @patch('%s.FccAnalysis.getEnvironmentScript' % MODULE_NAME, new=Mock(return_value=True))
  @patch("%s.FccAnalysis.generateBashScript" % MODULE_NAME, new=Mock(return_value=True))
  @patch("%s.glob.glob" % MODULE_NAME, new=Mock(return_value=[]))
  def test_runit_cardfile(self):
    card_file = "/path/to/cardFile"
    self.fccAna.randomGenerator = {"Pythia": [card_file]}
    self.fccAna.RandomSeed = 126
    self.fccAna.NumberOfEvents = 42

    with patch('%s.FccAnalysis.readFromFile' % MODULE_NAME) as mock_read, \
         patch('%s.FccAnalysis.writeToFile' % MODULE_NAME) as mock_write, \
         patch('os.makedirs'), \
         patch('os.path.exists') as mock_exists, \
         patch("%s.shellCall" % MODULE_NAME, new=Mock(return_value={'OK': True, 'Value': ["", "stdout", "stderr"]})) as mock_shellcall:

      mock_exists.side_effect = self.replace_exists
      content = "! CONTENT OF THE CARD FILE"
      message = 'Application : Card file reading successfull'
      mock_read.return_value = (content, message)
      mock_write.return_value = True

      assertDiracSucceedsWith(self.fccAna.runIt(), "Execution of the FCC application successfull", self)
      self.log_mock.info.assert_any_call("Environment : Environment script look up successfull")
      self.log_mock.debug.assert_any_call("Application code : Creation of the bash script successfull")
      self.log_mock.debug.assert_any_call("Application : Application execution and log file creation...")
      self.log_mock.debug.assert_any_call("Application : Application execution successfull")
      self.log_mock.debug.assert_any_call("Application : Log file creation successfull")
      self.log_mock.warn.assert_called_once_with(
          "Application : no root files have been generated, was that intentional ?")
      mock_shellcall.assert_called_once_with(
          0,
          self.fccAna.applicationScript,
          callbackFunction=self.fccAna.redirectLogOutput,
          bufferLimit=20971520)

      eventSetting = ["! N) AUTOMATIC GENERATION OF CODE DONE BY FCC APPLICATION FOR EVENT NUMBER SETTING"]
      eventSetting += ["Main:numberOfEvents = %d         ! number of events to generate" % self.fccAna.NumberOfEvents]
      contentWithEventSet = "%s\n%s\n" % (content, "\n".join(eventSetting))

      seedSetting = ["! N) AUTOMATIC GENERATION OF CODE DONE BY FCC APPLICATION FOR SEED NUMBER SETTING"]
      seedSetting += ["Random:setSeed = on         ! apply user-set seed everytime the Pythia::init is called"]
      seedSetting += ["Random:seed = %d         ! -1=default seed, 0=seed based on time, >0 user seed number" %
                      self.fccAna.RandomSeed]
      contentWithEventSeedSet = "%s\n%s\n" % (contentWithEventSet, "\n".join(seedSetting))

      self.log_mock.debug.assert_any_call(message)
      mock_read.assert_called_once_with(card_file)
      mock_write.assert_any_call('w', card_file, contentWithEventSeedSet)

  @patch('%s.FccAnalysis.getEnvironmentScript' % MODULE_NAME, new=Mock(return_value=True))
  @patch("%s.FccAnalysis.generateBashScript" % MODULE_NAME, new=Mock(return_value=True))
  @patch("%s.glob.glob" % MODULE_NAME, new=Mock(return_value=[]))
  def test_runit_cardfile_readfailed(self):
    card_file = "/path/to/cardFile"
    self.fccAna.randomGenerator = {"Pythia": [card_file]}
    self.fccAna.RandomSeed = 126
    self.fccAna.NumberOfEvents = 42

    with patch('%s.FccAnalysis.readFromFile' % MODULE_NAME) as mock_read, \
         patch('%s.FccAnalysis.writeToFile' % MODULE_NAME) as mock_write, \
         patch('os.path.exists') as mock_exists:

      mock_exists.side_effect = self.replace_exists
      message = 'Application : Card file reading failed'
      mock_read.return_value = (None, message)
      mock_write.return_value = True

      assertDiracFailsWith(self.fccAna.runIt(), message, self)

      self.log_mock.error.assert_called_once_with(message)
      mock_read.assert_called_once_with(card_file)

  @patch('%s.FccAnalysis.getEnvironmentScript' % MODULE_NAME, new=Mock(return_value=True))
  @patch("%s.FccAnalysis.generateBashScript" % MODULE_NAME, new=Mock(return_value=True))
  @patch("%s.glob.glob" % MODULE_NAME, new=Mock(return_value=[]))
  @patch("%s.FccAnalysis.writeToFile" % MODULE_NAME, new=Mock(return_value=False))
  def test_runit_cardfile_writefailed(self):
    card_file = "/path/to/cardFile"
    self.fccAna.randomGenerator = {"Pythia": [card_file]}
    self.fccAna.RandomSeed = 126
    self.fccAna.NumberOfEvents = 42

    with patch('%s.FccAnalysis.readFromFile' % MODULE_NAME) as mock_read, \
         patch('os.path.exists') as mock_exists:

      mock_exists.side_effect = self.replace_exists
      content = "******************"
      message = 'Application : Card file reading successfull'
      mock_read.return_value = (content, message)

      error_message = "Application : Card file overwitting failed"
      assertDiracFailsWith(self.fccAna.runIt(), error_message, self)

      self.log_mock.error.assert_called_once_with(error_message)
      mock_read.assert_called_once_with(card_file)

  def test_readfromfile(self):
    with patch('%s.open' % BUILTIN_NAME) as mock_open:
      manager = mock_open.return_value.__enter__.return_value
      manager.read.return_value = 'some data'
      content, message = self.fccAna.readFromFile("/my/file/to/read")
      assertEqualsImproved(content, 'some data', self)
      mock_open.assert_called_with("/my/file/to/read", 'r')
      assertEqualsImproved(message, 'Application : Card file reading successfull', self)

  @patch('%s.open' % BUILTIN_NAME, new=Mock(side_effect=IOError("ioerror")))
  def test_readfromfile_failed(self):
    content, message = self.fccAna.readFromFile("/my/file/to/read")
    assertEqualsImproved(None, content, self)
    assertEqualsImproved('Application : Card file reading failed\nioerror', message, self)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""
SLIC : Simulation after Whizard or StdHepCut
"""

from __future__ import absolute_import
import os
import types

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Workflow.Parameter import Parameter

from ILCDIRAC.Interfaces.API.NewInterface.LCApplication import LCApplication
from ILCDIRAC.Core.Utilities.InstalledFiles import Exists
import six

LOG = gLogger.getSubLogger(__name__)
__RCSID__ = "$Id$"


class SLIC(LCApplication):
  """Call SLIC simulator (after Whizard, Pythia or StdHepCut)

  Usage:

  >>> wh = Whizard()
  >>> slic = SLIC()
  >>> slic.getInputFromApp(wh)
  >>> slic.setSteeringFile("mymacrofile.mac")
  >>> slic.setStartFrom(10)

  Use :func:`setExtraCLIArguments` in case you want to use command line parameters
  """

  def __init__(self, paramdict=None):

    self.startFrom = 0
    self.randomSeed = 0
    self.detectorModel = ''
    super(SLIC, self).__init__(paramdict)
    # Those 5 need to come after default constructor
    self._modulename = 'SLICAnalysis'
    self._moduledescription = 'Module to run SLIC'
    self.appname = 'slic'
    self.datatype = 'SIM'
    self.detectortype = 'SID'
    self._paramsToExclude.extend(["outputDstPath", "outputRecPath", "OutputDstFile", "OutputRecFile"])

  def setRandomSeed(self, randomSeed):
    """ Optional: Define random seed to use. Default is 0.

    :param int randomSeed: Seed to use during simulation.
    """
    self._checkArgs({'randomSeed': int})

    self.randomSeed = randomSeed

  def setDetectorModel(self, detectorModel):
    """Define detector model to use for Slic simulation.

    :param str detectorModel: Detector Model to use for Slic simulation.
    """
    self._checkArgs({'detectorModel': (str,)})
    if detectorModel.lower().count("lfn:"):
      self.inputSB.append(detectorModel)
    elif detectorModel.lower().count(".zip"):
      if os.path.exists(detectorModel):
        self.inputSB.append(detectorModel)
      else:
        LOG.notice("Specified detector model does not exist locally, I hope you know what you're doing")

    self.detectorModel = os.path.basename(detectorModel).replace(".zip", "")

  def setStartFrom(self, startfrom):
    """ Optional: Define from where slic starts to read in the input file

    :param int startfrom: from where slic starts to read the input file
    """
    self._checkArgs({'startfrom': int})
    self.startFrom = startfrom

  def _userjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setUserJobFinalization(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('userjobmodules failed')
    return S_OK()

  def _prodjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setOutputComputeDataList(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('prodjobmodules failed')
    return S_OK()

  def _checkConsistency(self, job=None):

    if not self.version:
      return S_ERROR('No version found')
    if self.steeringFile:
      if not os.path.exists(self.steeringFile) and not self.steeringFile.lower().count("lfn:"):
        res = Exists(self.steeringFile)
        if not res['OK']:
          return res

    #res = self._checkRequiredApp()
    # if not res['OK']:
    #  return res

    if self._jobtype != 'User':
      self._listofoutput.append({"outputFile": "@{OutputFile}", "outputPath": "@{OutputPath}",
                                 "outputDataSE": '@{OutputSE}'})
      self.prodparameters['slic_steeringfile'] = self.steeringFile
      self.prodparameters['detectorType'] = self.detectortype
      if self.detectorModel:
        self.prodparameters['slic_detectormodel'] = self.detectorModel

    if not self.startFrom:
      LOG.info('No startFrom defined for Slic : start from the beginning')

    return S_OK()

  def _applicationModule(self):

    md1 = self._createModuleDefinition()
    md1.addParameter(Parameter("RandomSeed", 0, "int", "", "", False, False,
                               "Random seed for the generator"))
    md1.addParameter(Parameter("detectorModel", "", "string", "", "", False, False,
                               "Detecor model for simulation"))
    md1.addParameter(Parameter("startFrom", 0, "int", "", "", False, False,
                               "From how Slic start to read the input file"))
    md1.addParameter(Parameter("debug", False, "bool", "", "", False, False, "debug mode"))
    return md1

  def _applicationModuleValues(self, moduleinstance):

    moduleinstance.setValue("RandomSeed", self.randomSeed)
    moduleinstance.setValue("detectorModel", self.detectorModel)
    moduleinstance.setValue("startFrom", self.startFrom)
    moduleinstance.setValue("debug", self.debug)

  def _checkWorkflowConsistency(self):
    return self._checkRequiredApp()

  def _resolveLinkedStepParameters(self, stepinstance):
    if isinstance(self._linkedidx, six.integer_types):
      self._inputappstep = self._jobsteps[self._linkedidx]
    if self._inputappstep:
      stepinstance.setLink("InputFile", self._inputappstep.getType(), "OutputFile")
    return S_OK()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Tests for the Fcc module."""

from __future__ import absolute_import
import inspect
import os
import unittest
from mock import patch, MagicMock as Mock
import six

from DIRAC import S_OK, S_ERROR
from ILCDIRAC.Interfaces.API.NewInterface.Applications import FccSw, FccAnalysis

from Tests.Utilities.GeneralUtils import assertEqualsImproved, assertDiracFailsWith, \
    assertDiracSucceeds, assertDiracSucceedsWith

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Interfaces.API.NewInterface.Applications.Fcc'
BUILTIN_NAME = 'builtins' if six.PY3 else '__builtin__'

# pylint: disable=protected-access, too-many-public-methods, missing-docstring


class FccMixin(object):
  """Base class for the Fcc test cases."""

  def setUp(self):
    """set up the objects."""

    def replace_realpath(path):
      return os.path.join("/test/realpath", path)

    self.patches = [patch("os.getcwd", new=Mock(return_value="/test/curdir")),
                    patch("os.path.dirname", new=Mock(return_value="/test/dirname")),
                    patch("os.path.realpath", new=Mock(side_effect=replace_realpath)),
                    ]

    for patcher in self.patches:
      patcher.start()

    self.fcc = None
    self.log_mock = Mock()

  def tearDown(self):
    for patcher in self.patches:
      patcher.stop()
    del self.fcc

  def test_userjobmodules(self):
    module_mock = Mock()
    assertDiracSucceeds(self.fcc._userjobmodules(module_mock), self)

  def test_prodjobmodules(self):
    module_mock = Mock()
    assertDiracSucceeds(self.fcc._prodjobmodules(module_mock), self)

  def test_prodjobmodules_outputpath(self):
    module_mock = Mock()
    self.fcc.outputPath = 'aef'
    assertDiracSucceeds(self.fcc._prodjobmodules(module_mock), self)
    self.assertIn({'OutputFile': '@{OutputFile}', 'outputPath': '@{OutputPath}',
                     'outputDataSE': '@{OutputSE}'}, self.fcc._listofoutput)

  def test_userjobmodules_fails(self):
    with patch('%s._setUserJobFinalization' % MODULE_NAME, new=Mock(return_value=S_OK('something'))),\
        patch('%s._setApplicationModuleAndParameters' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_test_err'))):
      assertDiracFailsWith(self.fcc._userjobmodules(None),
                            'userjobmodules failed', self)

  def test_prodjobmodules_fails(self):
    with patch('%s._setApplicationModuleAndParameters' % MODULE_NAME, new=Mock(return_value=S_OK('something'))), \
        patch('%s._setOutputComputeDataList' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_other_test_err'))):
      assertDiracFailsWith(self.fcc._prodjobmodules(None),
                            'prodjobmodules failed', self)

  def test_checkwfconsistency(self):
    assertDiracSucceeds(self.fcc._checkWorkflowConsistency(), self)

  def test_resolveparameters(self):
    step_mock = Mock()
    assertDiracSucceeds(self.fcc._resolveLinkedStepParameters(step_mock()), self)

  def test_resolveparameters_setlink(self):
    step_mock = Mock()
    self.fcc._linkedidx = 1
    self.fcc._jobsteps = [None, Mock()]
    assertDiracSucceeds(self.fcc._resolveLinkedStepParameters(step_mock()), self)

  @patch("os.listdir", new=Mock(return_value=[]))
  @patch("os.path.exists", new=Mock(return_value=True))
  def test_checkconsistency(self):
    with patch.object(self.fcc, '_importToSandbox', new=Mock(return_value=True)), \
        patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):

      info_message = ("Application general consistency : _checkConsistency()"
                      " on '%(name)s' successfull" % {'name': self.fcc.appname}
                      )
      assertDiracSucceedsWith(self.fcc._checkConsistency(), info_message, self)
      self.log_mock.info.assert_called_with(info_message)
      self.log_mock.info.assert_any_call("Sandboxing : Sandboxing successfull")

  @patch("os.listdir", new=Mock(return_value=[]))
  @patch("os.path.exists", new=Mock(return_value=True))
  def test_checkconsistency_noversion(self):
    self.fcc.version = None
    error_message = 'Version not set!'
    with patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):
      assertDiracFailsWith(self.fcc._checkConsistency(), error_message, self)
    self.log_mock.error.assert_called_once_with(error_message)

  @patch("os.path.exists", new=Mock(return_value=True))
  def test_checkconsistency_noexecutable(self):
    self.fcc.fccExecutable = None
    error_message = (
        "Consistency : Error in parsing '%(name)s' application :\n"
        "You have to provide at least an executable"
        " and a configuration file for each application" % {'name': self.fcc.appname}
        )
    with patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):
      assertDiracFailsWith(self.fcc._checkConsistency(), error_message, self)
    self.log_mock.error.assert_called_once_with(error_message)

  @patch("os.listdir", new=Mock(return_value=[]))
  @patch("os.path.exists", new=Mock(return_value=True))
  def test_checkconsistency_no_cfgfile(self):
    self.fcc.steeringFile = None
    error_message = (
        "Consistency : Error in parsing '%(name)s' application :\n"
        "You have to provide at least an executable"
        " and a configuration file for each application" % {'name': self.fcc.appname}
        )
    with patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):
      assertDiracFailsWith(self.fcc._checkConsistency(), error_message, self)
    self.log_mock.error.assert_called_once_with(error_message)

  @patch("os.listdir", new=Mock(return_value=[]))
  @patch("os.path.exists", new=Mock(return_value=True))
  def test_checkconsistency_many_cfgfiles(self):
    self.fcc.steeringFile = ["/path/to/cfgFile1", "/path/to/cfgFile2"]
    error_message = (
        "Consistency : Fcc Application accepts only one input configuration file:\n"
        "If you want to run the application '%(name)s' with many configurations then\n"
        "Create an new application with the other configuration\n"
        "You can also use the 'getInputFromApp' method to link applications" % {'name': self.fcc.appname}
        )
    with patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):
      assertDiracFailsWith(self.fcc._checkConsistency(), error_message, self)
    self.log_mock.error.assert_called_once_with(error_message)

  def test_checkfinalconsistency_outputfile(self):
    self.fcc.logFile = "logFile"
    self.fcc.outputFile = "output.root"

    self.fcc._checkFinalConsistency()
    self.assertIn(self.fcc.logFile, self.fcc._outputSandbox)
    self.assertIn("output(_JobID).root (Name of the eventual output root file)", self.fcc._outputSandbox)

  # During release v0.8.1 of FCCSW, 'Application.setOutputFile' does not accept list
  def test_checkfinalconsistency_outputfiles(self):
    self.fcc.logFile = "logFile"
    self.fcc.outputFile = ["output1.root", "output2.root"]

    self.fcc._checkFinalConsistency()
    self.assertIn(self.fcc.logFile, self.fcc._outputSandbox)
    self.assertIn("output1(_JobID).root (Name of the eventual output root file)", self.fcc._outputSandbox)
    self.assertIn("output2(_JobID).root (Name of the eventual output root file)", self.fcc._outputSandbox)

  def test_importfiles_no_sandbox(self):
    self.fcc._tempInputSandbox = None
    with patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):
      self.assertTrue(self.fcc._importFiles())
    self.log_mock.warn.assert_called_once_with("Sandboxing : Your application has an empty input sandbox")

  @patch("%s._findPath" % MODULE_NAME, new=Mock(return_value=('sandboxFile1', False)))
  def test_importfiles_findpath_failed(self):
    self.fcc._tempInputSandbox = ['sandboxFile1']
    with patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):
      self.assertFalse(self.fcc._importFiles())
    error_message = (
        "Sandboxing : The path 'sandboxFile1' does not exist\n"
        "Please ensure that your path exists in an accessible file system "
        "(AFS or CVMFS)"
        )
    self.log_mock.error.assert_called_once_with(error_message)

  @patch("%s._findPath" % MODULE_NAME, new=Mock(return_value=('/afs/sandboxFile1', True)))
  def test_importfiles_afs_warn_check(self):
    self.fcc._tempInputSandbox = ['/afs/sandboxFile1']
    with patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):
      self.assertTrue(self.fcc._importFiles())
    warn_message = (
        "Sandboxing : You plan to upload '/afs/sandboxFile1'"
        " which is stored on AFS\n"
        "STORING FILES ON AFS IS DEPRECATED"
        )

    self.assertIn('/afs/sandboxFile1', self.fcc._inputSandbox)
    self.log_mock.warn.assert_called_once_with(warn_message)

    debug_message = (
        "Sandboxing : The path '/afs/sandboxFile1' required by the application"
        " has been added to te sandbox"
        )
    self.log_mock.debug.assert_any_call(debug_message)

    debug_message = (
        "Sandboxing : Files required by FCC application"
        " verified and added successfully to the sandbox"
        )
    self.log_mock.debug.assert_called_with(debug_message)

  def test_readfromfile(self):
    with patch('%s.open' % BUILTIN_NAME) as mock_open:
      manager = mock_open.return_value.__enter__.return_value
      manager.read.return_value = 'some data'
      content, message = self.fcc._readFromFile("/my/file/to/read")
      assertEqualsImproved(content, 'some data', self)
      mock_open.assert_called_with("/my/file/to/read", 'r')
      assertEqualsImproved(message, 'Sandboxing : FCC file reading successfull', self)

  @patch('%s.open' % BUILTIN_NAME, new=Mock(side_effect=IOError("ioerror")))
  def test_readfromfile_failed(self):
    content, message = self.fcc._readFromFile("/my/file/to/read")
    assertEqualsImproved(None, content, self)
    assertEqualsImproved('Sandboxing : FCC file reading failed\nioerror', message, self)

  @patch("os.listdir", new=Mock(return_value=[]))
  @patch('os.path.exists', new=Mock(return_value=True))
  @patch("%s._importToSandbox" % MODULE_NAME, new=Mock(return_value=False))
  def test_checkconsistency_importtosandbox_failed(self):
    assertDiracFailsWith(self.fcc._checkConsistency(), "_importToSandbox() failed", self)

  @patch('os.path.exists', new=Mock(return_value=True))
  def test_findPath(self):
    filename = '/path/to/file'
    assertEqualsImproved(self.fcc._findPath(filename), (filename, True), self)

  @patch('os.path.exists', new=Mock(return_value=False))
  def test_findPath_failed(self):
    filename = '/path/to/file'
    assertEqualsImproved(self.fcc._findPath(filename), (filename, False), self)


class FccSwTestCase(FccMixin, unittest.TestCase):
  """Tests for FccSw."""

  def setUp(self):
    super(FccSwTestCase, self).setUp()

    fcc_conf_file = '/build/YOUR_USERNAME/FCC/FCCSW/Examples/options/geant_fullsim_field.py'
    fccsw_path = '/build/YOUR_USERNAME/FCC/FCCSW'

    fccsw = FccSw(
        fccConfFile=fcc_conf_file,
        fccSwPath=fccsw_path
        )

    self.fcc = fccsw
    self.fcc._log = self.log_mock

  def test_checkconsistency_noexecutable(self):
    # NOTHING TO DO
    pass

  def test_checkconsistency_nofccswpath(self):
    self.fcc.fccSwPath = None
    error_message = (
        "FCCSW specific consistency : Error in parsing FCCSW application :\n"
        "You have to provide a valid path of the FCCSW installation"
        )
    with patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):
      assertDiracFailsWith(self.fcc._checkConsistency(), error_message, self)
    self.log_mock.error.assert_called_once_with(error_message)

  @patch('os.path.exists', new=Mock(return_value=False))
  def test_checkconsistency_exists_failed(self):
    error_message = (
        "FCCSW specific consistency : Error in parsing FCCSW application :\n"
        "You have to provide a valid path of the FCCSW installation"
        )
    with patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):
      assertDiracFailsWith(self.fcc._checkConsistency(), error_message, self)
    self.log_mock.error.assert_called_once_with(error_message)

  def test_checkconsistency_makedirs_failed(self):
    exists_dict = {self.fcc.fccSwPath: True, self.fcc._tempCwd: False}

    def replace_exists(path):
      return exists_dict[path]

    with patch('os.path.exists') as mock_exists, \
         patch('os.makedirs') as mock_makedirs, \
         patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):

      mock_exists.side_effect = replace_exists
      # Throw OS error
      mock_makedirs.side_effect = OSError("oserror")

      error_message = "FCCSW specific consistency : Creation of 'temp_fcc_dirac' folder failed\noserror"
      assertDiracFailsWith(self.fcc._checkConsistency(), error_message, self)
      self.log_mock.error.assert_called_once_with(error_message)
      mock_makedirs.assert_called_once_with(self.fcc._tempCwd)

  def test_checkconsistency_run_read(self):

    file_content = (
        '#!/bin/sh\n'
        'exec  /cvmfs/sft.cern.ch/lcg/views/LCG_88/x86_64-slc6-gcc49-opt/bin/python'
        ' /cvmfs/fcc.cern.ch/sw/0.8/gaudi/v28r2/x86_64-slc6-gcc49-opt/scripts/xenv '
        '--xml /afs/cern.ch/user/s/sfernana/FCC/FCCSW/build.x86_64-slc6-gcc49-opt/config/FCCSW-build.xenv "$@"\n'
        )

    build_folder = 'build.x86_64-slc6-gcc49-opt'
    build_path = os.path.join(self.fcc.fccSwPath, build_folder)
    executable_path = os.path.join(build_path, 'run')

    info_message = (
        "Application general consistency : _checkConsistency()"
        " on '%(name)s' successfull" % {'name': self.fcc.appname}
        )

    read_message = 'Sandboxing : FCC file reading successfull'

    with patch('os.path.exists') as mock_exists, \
         patch('os.listdir') as mock_listdir, \
         patch('os.path.isfile') as mock_isfile, \
         patch("%s._checkConsistency" % MODULE_NAME) as mock_check, \
         patch("%s._readFromFile" % MODULE_NAME) as mock_read, \
         patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):

      mock_exists.return_value = True
      mock_listdir.return_value = [build_folder]
      mock_isfile.return_value = False
      mock_read.return_value = (file_content, read_message)
      mock_check.return_value = S_OK(info_message)

      assertDiracSucceedsWith(self.fcc._checkConsistency(), info_message, self)
      self.log_mock.debug.assert_any_call(
          "FCCSW specific consistency : The temporary folder 'temp_fcc_dirac' already exists")
      mock_listdir.assert_called_once_with(self.fcc.fccSwPath)
      mock_isfile.assert_called_once_with(build_path)
      mock_exists.assert_any_call(executable_path)
      mock_read.assert_called_once_with(executable_path)
      self.log_mock.debug.assert_any_call(read_message)

  def test_checkconsistency_run_read_failed(self):

    build_folder = 'build.x86_64-slc6-gcc49-opt'
    build_path = os.path.join(self.fcc.fccSwPath, build_folder)
    executable_path = os.path.join(build_path, 'run')

    info_message = (
        "Application general consistency : _checkConsistency()"
        " on '%(name)s' successfull" % {'name': self.fcc.appname}
        )

    with patch('os.path.exists') as mock_exists, \
         patch('os.listdir') as mock_listdir, \
         patch('os.path.isfile') as mock_isfile, \
         patch("%s._checkConsistency" % MODULE_NAME) as mock_check, \
         patch("%s._readFromFile" % MODULE_NAME) as mock_read, \
         patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):

      mock_exists.return_value = True
      mock_listdir.return_value = [build_folder]
      mock_isfile.return_value = False
      mock_read.return_value = (None, 'DO NOT READ ME !')
      mock_check.return_value = S_OK(info_message)

      assertDiracSucceedsWith(self.fcc._checkConsistency(), info_message, self)
      self.log_mock.debug.assert_any_call(
          "FCCSW specific consistency : The temporary folder 'temp_fcc_dirac' already exists")
      mock_listdir.assert_called_once_with(self.fcc.fccSwPath)
      mock_isfile.assert_called_once_with(build_path)
      mock_exists.assert_any_call(executable_path)
      mock_read.assert_called_once_with(executable_path)
      self.log_mock.warn.assert_called_once_with('DO NOT READ ME !')
      self.log_mock.debug.assert_any_call(
          'FCCSW specific consistency : Using by default the command of the 0.8.1 release !')

  def test_resolvetreeoffiles(self):
    files = ['file1']
    source = os.path.realpath(os.path.join(self.fcc.fccSwPath, files[0]))
    destination = os.path.realpath(os.path.join(self.fcc._tempCwd, files[0]))

    tree = os.path.dirname(files[0])
    tree_full_path = os.path.join(self.fcc._tempCwd, tree)

    exists_dict = {tree_full_path: True, source: True, destination: False}

    def replace_exists(path):
      return exists_dict[path]

    with patch('os.path.exists') as mock_exists, \
         patch('shutil.copyfile') as mock_shutil, \
         patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):

      mock_exists.side_effect = replace_exists
      self.assertTrue(self.fcc._resolveTreeOfFiles(files, ".ext"))

      debug_message = "Sandboxing : Tree '%s' already exists" % tree_full_path
      self.log_mock.debug.assert_any_call(debug_message)

      debug_message = (
          "Sandboxing : Additionnal files"
          " '%(src)s' copy successfull to '%(dst)s'" % {'src': source, 'dst': destination}
          )
      self.log_mock.debug.assert_called_with(debug_message)
      mock_shutil.assert_called_once_with(source, destination)

  def test_resolvetreeoffiles_nofiles(self):
    files = []
    with patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):
      self.assertTrue(self.fcc._resolveTreeOfFiles(files, ".ext"))
    warn_message = (
        "Sandboxing : FCCSW configuration file"
        " does not seem to need any additional '.ext' files"
        )
    self.log_mock.warn.assert_called_once_with(warn_message)

  def test_resolvetreeoffiles_makedirs_failed(self):
    files = ['file1']
    source = os.path.realpath(os.path.join(self.fcc.fccSwPath, files[0]))

    tree = os.path.dirname(files[0])
    tree_full_path = os.path.join(self.fcc._tempCwd, tree)

    exists_dict = {tree_full_path: False, source: True}

    def replace_exists(path):
      return exists_dict[path]

    with patch('os.path.exists') as mock_exists, \
         patch('os.makedirs') as mock_makedirs, \
         patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):

      mock_makedirs.side_effect = OSError("oserror")
      mock_exists.side_effect = replace_exists

      self.assertFalse(self.fcc._resolveTreeOfFiles(files, ".ext"))
      error_message = (
          "Sandboxing : Tree '%(tree)s' of additionnal"
          " '.ext' files creation failed\noserror" % {'tree': tree_full_path}
          )
      self.log_mock.error.assert_called_once_with(error_message)
      mock_makedirs.assert_called_once_with(tree_full_path)

  @patch('os.path.exists', new=Mock(return_value=False))
  def test_resolvetreeoffiles_not_exists(self):
    files = ['file1']

    source = os.path.realpath(os.path.join(self.fcc.fccSwPath, files[0]))

    warn_message = (
        "Sandboxing : The file '%s' does not exist,"
        " it is not present in the FCCSW installation"
        "\nThen you should have added it manually to the input sandbox !" % {'source': source}
        )
    self.assertTrue(self.fcc._resolveTreeOfFiles(files, ".ext"))
    self.log_mock.warn(warn_message)

  def test_resolvetreeoffiles_shutil_failed(self):
    files = ['file1']

    tree = os.path.dirname(files[0])
    tree_full_path = os.path.join(self.fcc._tempCwd, tree)

    source = os.path.realpath(os.path.join(self.fcc.fccSwPath, files[0]))
    destination = os.path.realpath(os.path.join(self.fcc._tempCwd, files[0]))

    exists_dict = {source: True, tree_full_path: True, destination: False}

    def replace_exists(path):
      return exists_dict[path]

    with patch('os.path.exists') as mock_exists, \
         patch('os.makedirs'), \
         patch('shutil.copyfile') as mock_shutil, \
         patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):

      mock_exists.side_effect = replace_exists
      # throw OS error
      mock_shutil.side_effect = IOError("ioerror")

      self.assertFalse(self.fcc._resolveTreeOfFiles(files, ".ext"))

      error_message = (
          "Sandboxing : Additionnal files"
          " '%(src)s' copy failed\nioerror" % {'src': source}
          )

      self.log_mock.error.assert_called_once_with(error_message)
      mock_shutil.assert_called_once_with(source, destination)

  @patch("%s._importFiles" % MODULE_NAME, new=Mock(return_value=True))
  @patch("%s._importToSandbox" % MODULE_NAME, new=Mock(return_value=True))
  def test_importtosandbox(self):
    with patch.object(self.fcc, '_importFccswFiles', new=Mock(return_value=True)):
      self.assertTrue(self.fcc._importToSandbox())

  @patch("%s._importFiles" % MODULE_NAME, new=Mock(return_value=True))
  @patch("%s._importToSandbox" % MODULE_NAME, new=Mock(return_value=True))
  def test_importtosandbox_fccsw_files_failed(self):
    with patch.object(self.fcc, '_importFccswFiles', new=Mock(return_value=False)):
      self.assertFalse(self.fcc._importToSandbox())

  @patch("%s._importFiles" % MODULE_NAME, new=Mock(return_value=True))
  @patch("%s._importToSandbox" % MODULE_NAME, new=Mock(return_value=True))
  def test_importtosandbox_setfiltertofolders_failed(self):
    with patch.object(self.fcc, '_importFccswFiles', new=Mock(return_value=True)), \
        patch.object(self.fcc, '_setFilterToFolders', new=Mock(return_value=False)), \
        patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):

      self.assertFalse(self.fcc._importToSandbox())
      self.log_mock.error.assert_called_once_with("_setFilterToFolders() failed")

  @patch("%s._importFiles" % MODULE_NAME, new=Mock(return_value=True))
  @patch("%s._importToSandbox" % MODULE_NAME, new=Mock(return_value=False))
  def test_importtosandbox_super_method_failed(self):
    with patch.object(self.fcc, '_importFccswFiles', new=Mock(return_value=True)), \
        patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):
      self.assertFalse(self.fcc._importToSandbox())
      self.log_mock.error.assert_called_once_with("Sandboxing : _importToSandbox() failed")

  @patch("%s._readFromFile" % MODULE_NAME, new=Mock(return_value=('some content',
         'Sandboxing : FCC configuration file reading successfull')))
  def test_importfccswfiles(self):
    with patch.object(self.fcc, '_resolveTreeOfFiles') as mock_resolve:

      mock_resolve.return_value = True
      self.assertTrue(self.fcc._importFccswFiles())
      mock_resolve.assert_any_call([], '.txt')
      mock_resolve.assert_called_with([], '.cmd')

  def test_importfccswfiles_useof_pythia_generator(self):
    cmdFiles = ["Generation/data/Pythia_standard.cmd"]
    file_content = 'pythia8gentool = PythiaInterface("Pythia8Interface", Filename=pythiafile)\npythiafile="%s"' % cmdFiles[
        0]

    with patch.object(self.fcc, '_resolveTreeOfFiles') as mock_resolve, \
         patch("%s._readFromFile" % MODULE_NAME) as mock_read:

      mock_resolve.return_value = True
      mock_read.return_value = (file_content, 'Sandboxing : FCC file reading successfull')
      self.assertTrue(self.fcc._importFccswFiles())
      assertEqualsImproved(self.fcc.randomGenerator["Pythia"], cmdFiles, self)
      mock_resolve.assert_called_with(cmdFiles, '.cmd')
      mock_resolve.assert_any_call([], '.txt')

  def test_importfccswfiles_useof_gaudi_generator(self):
    cmdFiles = ["Generation/data/Pythia_standard.cmd"]
    file_content = 'Pythia is not used, it is commented\n#pythia8gentool = PythiaInterface("Pythia8Interface", Filename=pythiafile)\npythiafile="%s"\nGaudi ParticleGun somewhere' % cmdFiles[
        0]

    with patch.object(self.fcc, '_resolveTreeOfFiles') as mock_resolve, \
         patch("%s._readFromFile" % MODULE_NAME) as mock_read:

      mock_resolve.return_value = True
      mock_read.return_value = (file_content, 'Sandboxing : FCC configuration file reading successfull')
      self.assertTrue(self.fcc._importFccswFiles())
      self.assertTrue(self.fcc.randomGenerator["Gaudi"])
      mock_resolve.assert_called_with(cmdFiles, '.cmd')
      mock_resolve.assert_any_call([], '.txt')

  @patch("%s._readFromFile" % MODULE_NAME, new=Mock(return_value=("", "error message")))
  def test_importfccswfiles_read_failed(self):
    with patch.object(self.fcc, '_resolveTreeOfFiles', new=Mock(return_value=True)), \
        patch("os.path.exists", new=Mock(return_value=True)), \
        patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):

      self.assertFalse(self.fcc._importFccswFiles())
      self.log_mock.error.assert_called_once_with("error message")
      detectorFolder = os.path.join(self.fcc.fccSwPath, 'Detector')
      self.assertIn(detectorFolder, self.fcc._foldersToFilter)

  @patch("os.path.exists", new=Mock(return_value=False))
  @patch("%s._readFromFile" % MODULE_NAME, new=Mock(return_value=("some content", "some message")))
  def test_importfccswfiles_detector_warn(self):
    with patch.object(self.fcc, '_resolveTreeOfFiles', new=Mock(return_value=True)):

      self.assertTrue(self.fcc._importFccswFiles())

      warn_message = (
          "Sandboxing : The folder 'Detector' does not exist,"
          " it is not present in the FCCSW installation"
          "\nThen you should have added it manually to the input sandbox !"
          )
      self.log_mock.warn(warn_message)

      detectorFolder = os.path.join(self.fcc.fccSwPath, 'Detector')
      self.assertNotIn(detectorFolder, self.fcc._foldersToFilter)

  @patch("%s._readFromFile" % MODULE_NAME, new=Mock(return_value=("some content", "some message")))
  def test_importfccswfiles_resolvefiles_failed(self):
    with patch.object(self.fcc, '_resolveTreeOfFiles', new=Mock(return_value=False)), \
        patch("os.path.exists", new=Mock(return_value=True)), \
        patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):

      self.assertFalse(self.fcc._importFccswFiles())
      self.log_mock.error.assert_called_once_with("Sandboxing : _resolveTreeOfFiles() failed")
      detectorFolder = os.path.join(self.fcc.fccSwPath, 'Detector')
      self.assertIn(detectorFolder, self.fcc._foldersToFilter)

  def test_checkfinalconsistency(self):
    self.fcc._checkFinalConsistency()
    self.assertTrue(self.fcc.isGaudiOptionsFileNeeded)

  @patch("os.path.exists", new=Mock(return_value=False))
  def test_setfiltertofolders_exists_failed(self):
    self.fcc._foldersToFilter = set(['folder_to_filter1'])
    with patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):
      self.assertFalse(self.fcc._setFilterToFolders())
    error_message = (
        "Sandboxing : _filterFolders() failed\n"
        "The folder '%(folder)s' does not exist\n"
        "Check if you're FCCSW installation is complete"
        ) % {'folder': self.fcc._foldersToFilter.pop()}
    self.log_mock.error.assert_called_once_with(error_message)

  @patch("os.path.exists", new=Mock(return_value=True))
  def test_setfiltertofolders_filtering_failed(self):
    with patch.object(self.fcc, '_filterFolders', new=Mock(return_value=False)), \
        patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):
      self.fcc._foldersToFilter = set(['folder_to_filter1'])
      self.assertFalse(self.fcc._setFilterToFolders())
      self.log_mock.error.assert_called_once_with("Sandboxing : _filterFolders() failed")

  @patch("os.path.exists", new=Mock(return_value=True))
  def test_setfiltertofolders_filtering_succeed(self):
    with patch.object(self.fcc, '_filterFolders', new=Mock(return_value=True)), \
        patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):
      self.fcc._foldersToFilter = set(['folder_to_filter1'])
      self.assertTrue(self.fcc._setFilterToFolders())
      self.log_mock.debug.assert_called_with("Sandboxing : Folders filtering successfull")
      temp_folder = os.path.join(self.fcc._tempCwd, os.path.basename('folder_to_filter1'))
      self.assertIn(temp_folder, self.fcc._foldersToFilter)

  def test_filterfolders_recursivity(self):
    temp_folder = "/my/temp/folder"
    actual_folder = "/my/actual/folder"
    actual_sub_folder = os.path.join(actual_folder, 'folder2')

    listdir_dict = {actual_folder: ['file11', 'file12', 'folder2'], actual_sub_folder: ['file21', 'file22']}

    def replace_listdir(path):
      return listdir_dict[path]

    isfile_dict = {'file11': True, 'file12': True, 'folder2': False, 'file21': True, 'file22': True}

    def replace_isfile(path):
      return isfile_dict[os.path.basename(path)]

    with patch('os.path.exists') as mock_exists, \
         patch('os.listdir') as mock_listdir, \
         patch('os.makedirs'), \
         patch('shutil.copyfile'), \
         patch('os.path.isfile') as mock_isfile, \
         patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):

      mock_exists.return_value = True

      mock_listdir.side_effect = replace_listdir
      mock_isfile.side_effect = replace_isfile

      self.assertTrue(self.fcc._filterFolders(temp_folder, actual_folder, ".ext", False))

      mock_exists.assert_any_call(temp_folder)

      mock_listdir.assert_any_call(actual_folder)
      mock_listdir.assert_any_call(actual_sub_folder)

      actual_folder = os.path.realpath(actual_folder)
      actual_sub_folder = os.path.realpath(actual_sub_folder)

      mock_isfile.assert_any_call(os.path.join(actual_folder, 'file11'))
      mock_isfile.assert_any_call(os.path.join(actual_folder, 'file12'))
      mock_isfile.assert_any_call(os.path.join(actual_folder, 'folder2'))
      mock_isfile.assert_any_call(os.path.join(actual_sub_folder, 'file21'))
      mock_isfile.assert_any_call(os.path.join(actual_sub_folder, 'file22'))

      mock_exists.assert_any_call(os.path.join(actual_folder, 'file11'))
      mock_exists.assert_any_call(os.path.join(actual_folder, 'file12'))
      mock_exists.assert_any_call(os.path.join(actual_folder, 'folder2'))
      mock_exists.assert_any_call(os.path.join(actual_sub_folder, 'file21'))
      mock_exists.assert_any_call(os.path.join(actual_sub_folder, 'file22'))

      debug_message = "Sandboxing : Folder '%s' filtering successfull" % temp_folder
      self.log_mock.debug.assert_any_call(debug_message)

  def test_filterfolders_exists_failed(self):
    temp_folder = "/my/temp/folder"
    actual_folder = "/my/actual/folder"
    actual_sub_folder = os.path.join(actual_folder, 'folder2')

    listdir_dict = {actual_folder: ['file11', 'file12', 'folder2'], actual_sub_folder: ['file21', 'file22']}

    def replace_listdir(path):
      return listdir_dict[path]

    with patch('os.path.exists') as mock_exists, \
         patch('os.listdir') as mock_listdir, \
         patch('os.makedirs') as mock_makedirs, \
         patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):

      mock_exists.return_value = False

      mock_listdir.side_effect = replace_listdir

      self.assertFalse(self.fcc._filterFolders(temp_folder, actual_folder, ".ext", False))

      mock_exists.assert_any_call(temp_folder)
      mock_makedirs.assert_called_once_with(temp_folder)

      mock_listdir.assert_called_once_with(actual_folder)

      debug_message = (
          "Sandboxing : Creation of the filtered folder"
          " '%(temp)s' successfull" % {'temp': temp_folder}
          )

      self.log_mock.debug.assert_any_call(debug_message)

      error_message = "Sandboxing : The file '%s' does not exist" % os.path.realpath(
          os.path.join(actual_folder, 'file11'))
      self.log_mock.error.assert_called_once_with(error_message)

  @patch('os.path.exists', new=Mock(return_value=False))
  def test_filterfolders_makedirs_failed(self):
    temp_folder = "/my/temp/folder"
    actual_folder = "/my/actual/folder"

    with patch('os.makedirs') as mock_makedirs, \
         patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):

      mock_makedirs.side_effect = OSError("oserror")

      self.assertFalse(self.fcc._filterFolders(temp_folder, actual_folder, ".ext", False))

      error_message = (
          "Sandboxing : Creation of the filtered folder"
          " '%(temp)s' failed\noserror" % {'temp': temp_folder}
          )
      self.log_mock.error.assert_called_once_with(error_message)

      mock_makedirs.assert_called_once_with(temp_folder)

  def test_filterfolders_shutil_failed(self):
    temp_folder = "/my/temp/folder"
    actual_folder = "/my/actual/folder"

    source = os.path.realpath(os.path.join(actual_folder, 'file11'))
    destination = os.path.realpath(os.path.join(temp_folder, 'file11'))

    exists_dict = {temp_folder: True, source: True, destination: False}

    def replace_exists(path):
      return exists_dict[path]

    listdir_dict = {actual_folder: ['file11']}

    def replace_listdir(path):
      return listdir_dict[path]

    isfile_dict = {'file11': True}

    def replace_isfile(path):
      return isfile_dict[os.path.basename(path)]

    with patch('os.path.exists') as mock_exists, \
         patch('os.listdir') as mock_listdir, \
         patch('shutil.copyfile') as mock_shutil, \
         patch('os.path.isfile') as mock_isfile, \
         patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):

      # throw OS error
      mock_shutil.side_effect = IOError("ioerror")

      mock_listdir.side_effect = replace_listdir
      mock_isfile.side_effect = replace_isfile
      mock_exists.side_effect = replace_exists

      self.assertTrue(self.fcc._filterFolders(temp_folder, actual_folder, ".ext", True))

      mock_exists.assert_any_call(temp_folder)
      mock_exists.assert_any_call(source)
      mock_exists.assert_any_call(destination)

      mock_listdir.assert_called_once_with(actual_folder)

      mock_isfile.assert_called_once_with(os.path.realpath(os.path.join(actual_folder, 'file11')))

      mock_shutil.assert_called_once_with(source, destination)

      warn_message = "Sandboxing : The copy of the file '%s' failed\nioerror" % destination
      self.log_mock.warn.assert_called_once_with(warn_message)

  def test_filterfolders_exclude_txt(self):
    temp_folder = "/my/temp/folder"
    actual_folder = "/my/actual/folder"

    source1 = os.path.realpath(os.path.join(actual_folder, 'file11.txt'))
    destination1 = os.path.realpath(os.path.join(temp_folder, 'file11.txt'))

    source2 = os.path.realpath(os.path.join(actual_folder, 'file12.zip'))
    destination2 = os.path.realpath(os.path.join(temp_folder, 'file12.zip'))

    exists_dict = {temp_folder: True, source1: True, destination1: False, source2: True, destination2: False}

    def replace_exists(path):
      return exists_dict[path]

    listdir_dict = {actual_folder: ['file11.txt', 'file12.zip']}

    def replace_listdir(path):
      return listdir_dict[path]

    isfile_dict = {'file11.txt': True, 'file12.zip': True}

    def replace_isfile(path):
      return isfile_dict[os.path.basename(path)]

    with patch('os.path.exists') as mock_exists, \
         patch('os.listdir') as mock_listdir, \
         patch('shutil.copyfile') as mock_shutil, \
         patch('os.path.isfile') as mock_isfile, \
         patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):

      mock_listdir.side_effect = replace_listdir
      mock_isfile.side_effect = replace_isfile
      mock_exists.side_effect = replace_exists

      # exclude all txt files
      self.assertTrue(self.fcc._filterFolders(temp_folder, actual_folder, ".txt", True))

      mock_exists.assert_any_call(temp_folder)
      mock_exists.assert_any_call(source1)
      mock_exists.assert_any_call(destination1)

      mock_exists.assert_any_call(source2)
      mock_exists.assert_any_call(destination2)

      mock_listdir.assert_called_once_with(actual_folder)

      mock_isfile.assert_any_call(os.path.realpath(os.path.join(actual_folder, 'file11.txt')))

      # zip files are copied so shutil called for these files
      mock_shutil.assert_any_call(source2, destination2)

      # txt files are not copied (they are excluded) so shutil not called for these files
      self.assertFalse((source1, destination1) in mock_shutil.call_args_list)

      debug_message = "Sandboxing : Folder '%s' filtering successfull" % temp_folder
      self.log_mock.debug.assert_called_with(debug_message)

  def test_filterfolders_include_txt(self):
    temp_folder = "/my/temp/folder"
    actual_folder = "/my/actual/folder"

    source1 = os.path.realpath(os.path.join(actual_folder, 'file11.txt'))
    destination1 = os.path.realpath(os.path.join(temp_folder, 'file11.txt'))

    source2 = os.path.realpath(os.path.join(actual_folder, 'file12.zip'))
    destination2 = os.path.realpath(os.path.join(temp_folder, 'file12.zip'))

    exists_dict = {temp_folder: True, source1: True, destination1: False, source2: True, destination2: False}

    def replace_exists(path):
      return exists_dict[path]

    listdir_dict = {actual_folder: ['file11.txt', 'file12.zip']}

    def replace_listdir(path):
      return listdir_dict[path]

    isfile_dict = {'file11.txt': True, 'file12.zip': True}

    def replace_isfile(path):
      return isfile_dict[os.path.basename(path)]

    with patch('os.path.exists') as mock_exists, \
         patch('os.listdir') as mock_listdir, \
         patch('shutil.copyfile') as mock_shutil, \
         patch('os.path.isfile') as mock_isfile, \
         patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):

      mock_listdir.side_effect = replace_listdir
      mock_isfile.side_effect = replace_isfile
      mock_exists.side_effect = replace_exists

      # include only txt files
      self.assertTrue(self.fcc._filterFolders(temp_folder, actual_folder, ".txt", False))

      mock_exists.assert_any_call(temp_folder)
      mock_exists.assert_any_call(source1)
      mock_exists.assert_any_call(destination1)

      mock_exists.assert_any_call(source2)
      mock_exists.assert_any_call(destination2)

      mock_listdir.assert_called_once_with(actual_folder)

      mock_isfile.assert_any_call(os.path.realpath(os.path.join(actual_folder, 'file11.txt')))

      # txt files are copied so shutil called for these files
      mock_shutil.assert_any_call(source1, destination1)

      # zip files are not copied (they are excluded) so shutil not called for these files
      self.assertFalse((source2, destination2) in mock_shutil.call_args_list)

      debug_message = "Sandboxing : Folder '%s' filtering successfull" % temp_folder
      self.log_mock.debug.assert_called_with(debug_message)


class FccAnalysisTestCase(FccMixin, unittest.TestCase):
  """Tests for FccAnalysis."""

  def setUp(self):

    super(FccAnalysisTestCase, self).setUp()

    fcc_conf_file = '/path/to/ee_ZH_Zmumu_Hbb.txt'
    output_file = "ee_ZH_Zmumu_Hbb.root"

    fccphysics = FccAnalysis(
        fccConfFile=fcc_conf_file
        )

    fccphysics.setOutputFile(output_file)

    self.fcc = fccphysics

  def test_randomGenerator(self):
    assertEqualsImproved(self.fcc.randomGenerator, {"Pythia": []}, self)

  def test_readeventfalse(self):
    self.assertFalse(self.fcc.read)

  def test_readeventtrue(self):
    fccphysics_read = FccAnalysis(
        executable='fcc-physics-read',
        fccConfFile="/path/to/confFile"
        )
    self.assertTrue(fccphysics_read.read)

    fccphysics_read_delphes = FccAnalysis(
        executable='fcc-physics-read-delphes',
        fccConfFile="/path/to/confFile"
        )
    self.assertTrue(fccphysics_read_delphes.read)

  @patch("%s._importFiles" % MODULE_NAME, new=Mock(return_value=True))
  def test_importtosandbox(self):
    with patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):
      self.assertTrue(self.fcc._importToSandbox())
    self.log_mock.debug.assert_called_with("Sandboxing : Importation of user files/folders successfull")

  @patch("%s._importFiles" % MODULE_NAME, new=Mock(return_value=False))
  def test_importtosandbox_failed(self):
    with patch.object(inspect.getmodule(FccSw), 'LOG', new=self.log_mock):
      self.assertFalse(self.fcc._importToSandbox())
    self.log_mock.error.assert_called_once_with("Sandboxing : _importFiles() failed")



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Mixin for DD4hep interface functions."""

from __future__ import absolute_import
import types
import os

from DIRAC import S_OK, S_ERROR, gLogger

LOG = gLogger.getSubLogger("ILCJob")


class DDInterfaceMixin(object):
  """Mixin for DD4hep interface functions.

  This mixin class requires that the main class should be of
  :class:`~ILCDIRAC.Interfaces.API.NewInterface.Application.Application` type and have the
  ``detectorModel``, ``_ops``, ``version``, and ``_log`` attributes.

   * ``_ops`` is an Operations instance
   * ``_log`` a gLogger sublogger
   * ``version`` is the version of the application
   * ``detectorModel`` is set to the name of the detector model as inferred in the :func:`setDetectorModel` function
  """

  def setDetectorModel(self, detectorModel):
    """Define detector model to use for ddsim simulation.

    The detector model can be a collection of XML files. Either one has to use a
    detector model provided by LCGeo or DD4hep, which can be found on CVMFS or
    the complete XML needs to be passed as a tarball in the input sandbox or on the grid

    >>> ddsim.setDetectorModel("CLIC_o3_v13") # detector model part of lcgeo

    >>> ddsim.setDetectorModel("CLIC_o3_v13.tar.gz") # local tarball, will be added to input sandbox

    >>> ddsim.setDetectorModel("LFN:/ilc/user/u/username/CLIC_o3_v13.tar.gz") # tarball on the grid

    The tarball name must be detectorModel plus extension.
    The tarball must contain all xml files inside a folder called detectorModel.
    That is the main file is located in *detectorModel/detectorModel.xml*

    :param str detectorModel: Detector Model to use for simulation or reconstruction. Can
      be on CVMFS, tarball LFN or inputSandbox tarball
    """
    self._checkArgs({'detectorModel': (str,)})
    extensions = (".zip", ".tar.gz", ".tgz")

    # file on the grid
    if detectorModel.lower().startswith("lfn:"):
      self.inputSB.append(detectorModel)
      self.detectorModel = os.path.basename(detectorModel)
      for ext in extensions:
        if detectorModel.endswith(ext):
          self.detectorModel = os.path.basename(detectorModel).replace(ext, '')
      return S_OK()

    # local file
    elif detectorModel.endswith(extensions):
      for ext in extensions:
        if detectorModel.endswith(ext):
          self.detectorModel = os.path.basename(detectorModel).replace(ext, '')
          break

      if os.path.exists(detectorModel):
        self.inputSB.append(detectorModel)
      else:
        LOG.notice("Specified detector model file does not exist locally, I hope you know what you're doing")
      return S_OK()

    # DetectorModel is part of the software
    else:
      knownDetectors = self.getKnownDetectorModels()
      if not knownDetectors['OK']:
        return self._reportError("Failed to get knownDetectorModels: %s" % knownDetectors["Message"])
      elif detectorModel in knownDetectors['Value']:
        self.detectorModel = detectorModel
      else:
        return self._reportError("Unknown detector model in %s: %s" % (self.appname, detectorModel))
    return S_OK()

  def getKnownDetectorModels(self, version=None):
    """return a list of known detectorModels.

    Depends on the version of the software though...

    :param str version: Optional: Software version for which to print the detector models. If not given the version of the application instance is used.
    :returns: S_OK with list of detector models known for this software version, S_ERROR
    """
    if version is None and not self.version:
      return S_ERROR("No software version defined")
    detectorModels = self._ops.getOptionsDict("/DDSimDetectorModels/%s" % (self.version))
    return detectorModels
